#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/shm.h>
#include <semaphore.h>
#include <sys/sem.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <stdbool.h>
#include <sys/ioctl.h>
#include <math.h>
#include <signal.h>
#include <sys/syscall.h>
#include <pthread.h>

#include "EcatDeviceAPI.h"
#include "EcatErrno.h"
#include "EcatGlobals.h"
#include "Slave.h"
#include "ixecat.h"
#include "config.h"


// bar0 reg offset
#define GPIO_ADDR_OFFSET 						0x000
#define CARD_ID_ADDR_OFFSET					0x004
#define CMP_TRIG_COMM_CONFIG_ADDR_OFFSET 0x00C
#define ENC0_MODE_ADDR_OFFSET 	0x010
#define ENC0_CNT_ADDR_OFFSET 		0x014
#define ENC1_MODE_ADDR_OFFSET 	0x018
#define ENC1_CNT_ADDR_OFFSET 		0x01C
#define ENC0_CMP_TRIG_DATA_ADDR_OFFSET 		0x020
#define ENC0_CMP_TRIG_PITCH_ADDR_OFFSET 		0x024
#define ENC0_CMP_TRIG_TIMES_ADDR_OFFSET 		0x028
#define ENC0_CMP_TRIG_CONFIG_ADDR_OFFSET 		0x02C
#define ENC1_CMP_TRIG_DATA_ADDR_OFFSET 		0x034
#define ENC1_CMP_TRIG_PITCH_ADDR_OFFSET 		0x038
#define ENC1_CMP_TRIG_TIMES_ADDR_OFFSET 		0x03C
#define ENC1_CMP_TRIG_CONFIG_ADDR_OFFSET 		0x040

// bar2 reg offset
#define POLLING_FLAG_ADDR_OFFSET 0x00
#define CMD_RET_ADDR_OFFSET 0x01
#define DATA_ADDR_OFFSET 0x02
#define PC_INT_RDC_ADDR_OFFSET 0x7FF

#define DEV_ENCODER_MAX 2
#define MC_MPG_AXIS_NO_MAX                   	7

#define MC_DOUBLE_EQUAL_ZERO			0.00001

#define M_PI		3.14159265358979323846	/* pi */
#define M_2PI		6.283185307179586476925286766559

#define MC_CIRC_MODE_CP_ANGLE		0  //Center Point, Angle
#define MC_CIRC_MODE_CP_EP      1  //Center Point, End Point
#define MC_CIRC_MODE_BP_EP      2  //Border Point, End Point

#define DIFF_NS(A, B) 						(((A).tv_sec - (B).tv_sec) * NSEC_PER_SEC + (A).tv_nsec - (B).tv_nsec)
#define CHECK_ECAT_SLAVE_NO(a) if(a >= NETWORK_SLAVE_MAX) return ECAT_ERR_INVALID_SLAVE_NO;

typedef struct stewartPlatform {
	double rb;
	double rp;
	double ab;
	double ap;
	double ab_degree;
	double ap_degree;
	double L0[6];
	double Z0;
	double angP[7];
	double Px[7];
	double Py[7];
	double Pz[7];
	double angB[7];
	double Bx[7];
	double By[7];
	double Bz[7];
	double Lmin[6];
	double Lmax[6];
	int setup;
}stewartPlatform_t;

typedef struct ecat_device_ipc {
	int Fd;
	bool IsOpen;
	sem_t	ThreadSem;
	uint16_t GrpAxisCnt[MC_GROUP_NO_MAX];
	uint16_t	MpgAxisNo[MC_MPG_AXIS_NO_MAX];
	uint16_t MpgAxisCnt;
	pthread_t MpgThreadID;
	uint8_t MpgThreadRunning;
	uint8_t MpgThreadShouldStop;
	stewartPlatform_t	StewartPlatform;
	uint8_t irq_enable;
	struct sigaction sig_act;
	pthread_cond_t  cond_timer;
	pthread_mutex_t cond_timer_mutex;
	pthread_cond_t  cond_event;
	pthread_mutex_t cond_event_mutex;
	ecat_slave_t Slave[NETWORK_SLAVE_MAX];
}ecat_device_ipc_t;
/*****************************************************************************/
typedef union encoder_config {
	struct
	{
		uint8_t	 Mode0 : 1;
		uint8_t	 EncoderMode : 2;
		uint8_t	 NC3 : 1;
		uint8_t	 LPA1 : 1;
		uint8_t	 LPB1 : 1;
		uint8_t	 NC6_7 : 2;
		uint8_t	 LPF : 8;
		uint16_t NC16_31 : 16;
	};
	uint32_t Cfg;
}encoder_config_t;



const double DegtoRad = 0.01745329251;//   pi/180

static ecat_device_ipc_t EcatDevice[CARD_DEVICE_NO_MAX] = {{0}};

#define CHECK_ECAT_DEVICE_NO(a) if(a >= CARD_DEVICE_NO_MAX) return ECAT_ERR_IPC_INVALID_DEVICE_NO;
#define CHECK_ECAT_DEVICE_IS_OPEN(a) if(!a) return ECAT_ERR_IPC_DEVICE_NOT_OPEN;
#define CHECK_MC_PROFILE_NO(a)				if(a >= MC_PROFILE_NO_MAX) return ECAT_ERR_MC_INVALID_PROFILE_NO;
#define CHECK_MC_ECAM_TABLE_NO(a)				if(a >= MC_ECAM_TABLE_NO_MAX) return ECAT_ERR_MC_INVALID_ECAM_TABLE_NO;
#define SEMEPHORE_P_OPERATION(a) sem_wait(&a);
#define SEMEPHORE_V_OPERATION(a) sem_post(&a);
#define CHECK_PID_NO(a) if(a>=ECAT_ERR_INVALID_PID_NO || a<0) return ECAT_ERR_INVALID_PID_NO;

static struct timespec polling_time = { .tv_sec = 0 , .tv_nsec = ECAT_IPC_POLLING_NS};

static int32_t SetCommandData(ecat_device_ipc_t *EcatDevice, int32_t CmdNo, void *Data, uint32_t Size);
static int32_t SetCommandData_OpenDevice(ecat_device_ipc_t *EcatDevice, int32_t CmdNo, void *Data, uint32_t Size);
static int32_t SetCommandDataEx(ecat_device_ipc_t *EcatDevice, int32_t CmdNo, void *Data, uint32_t Size, uint64_t TimeOut_ms);
static int32_t GetReturnData(ecat_device_ipc_t *EcatDevice, void *Data, uint32_t Size);
static int32_t GenericSendCommand(uint16_t device_no, uint64_t timeout_ms, int32_t cmd_no,
									void *args, uint32_t args_size, void *ret_data, uint32_t ret_size);

static void Vector_Cross(uint16_t Dim, double *OutVector, double *Vect1, double *Vect2);
static double Vector_Dot(uint16_t Dim, double *Vect1, double *Vect2);
static int32_t CheckNormalVector(double *Vec);
static int32_t FindNV(double *SP, double *CP, double *EP, double *NV);
static int32_t FindCP_2D(double *p1, double *p2, double *p3, double *CP);
static int32_t FindCP_3D(double *p1, double *p2, double *p3, double *CP, double *NV);
static int32_t InverseMatrix(double *a);

/*****************************************************************************/
static void timespec_add_nsec(struct timespec *r, const struct timespec *a, int64_t b)
{
	r->tv_sec = a->tv_sec + (b / NSEC_PER_SEC);
	r->tv_nsec = a->tv_nsec + (b % NSEC_PER_SEC);

	if (r->tv_nsec >= NSEC_PER_SEC) {
		r->tv_sec++;
		r->tv_nsec -= NSEC_PER_SEC;
	} else if (r->tv_nsec <= -NSEC_PER_SEC) {
		r->tv_sec--;
		r->tv_nsec += NSEC_PER_SEC;
	}
}
/*****************************************************************************/
int32_t read_bar0_reg(int fd, uint32_t offset, uint32_t *data, uint32_t size)
{
	uint32_t i;
	ixecat_reg_t reg;
	reg.id = IXECAT_HOST_B0;

	for (i = 0; i < size; i++) {
		reg.reg_off = offset + (i * 0x04);
		if (ioctl(fd, IXECAT_READ_REG, &reg))
			return ECAT_ERR_IPC_READ_SHM;
		else
			data[i] = reg.value;
	}
	return 0;

}
/*****************************************************************************/
int32_t write_bar0_reg(int fd, uint32_t offset, uint32_t *data, uint32_t size)
{
	uint32_t i;
	ixecat_reg_t reg;
	reg.id = IXECAT_HOST_B0;

	for (i = 0; i < size; i++) {
		reg.reg_off = offset + (i * 0x04);
		reg.value = data[i];
		if (ioctl(fd, IXECAT_WRITE_REG, &reg))
			return ECAT_ERR_IPC_WRITE_SHM;
	}
	return 0;

}
/*****************************************************************************/
int32_t read_dpram(int fd, uint32_t offset, uint32_t *data, uint32_t size)
{
	uint32_t i;
	ixecat_reg_t reg;
	reg.id = IXECAT_SRAM1;

	for (i = 0; i < size; i++) {
		reg.sram_off = offset + i;
		if (ioctl(fd, IXECAT_READ_REG, &reg))
			return ECAT_ERR_IPC_READ_SHM;
		else
			data[i] = reg.value;
	}
	return 0;

}
/*****************************************************************************/
int32_t write_dpram(int fd, uint32_t offset, uint32_t *data, uint32_t size)
{
	uint32_t i;
	ixecat_reg_t reg;
	reg.id = IXECAT_SRAM1;

	for (i = 0; i < size; i++) {
		reg.sram_off = offset + i;
		reg.value = data[i];
		if (ioctl(fd, IXECAT_WRITE_REG, &reg))
			return ECAT_ERR_IPC_WRITE_SHM;
	}
	return 0;

}
/*****************************************************************************/
int32_t __SetCommandData(int fd, int32_t cmd_no, uint32_t *data,uint32_t size, uint32_t flag_value, uint64_t timeout_ms)
{
	struct timespec start_time, end_time;
	uint32_t resize = size / ECAT_IPC_SHM_ALIGN;
	uint32_t flag;
	uint64_t timeout_ns;
	if (timeout_ms == 0)
		timeout_ns = ECAT_IPC_TIMEOUT_NS;
	else
		timeout_ns = timeout_ms * NSEC_PER_MSEC;


	//get polling flag
	read_dpram(fd, POLLING_FLAG_ADDR_OFFSET, &flag, 1);
	if (flag != 0)
		return ECAT_ERR_IPC_BUSY;

	//set command No.
	write_dpram(fd, CMD_RET_ADDR_OFFSET, &cmd_no, 1);

	//set oprations data
	if(data != NULL)
		write_dpram(fd, DATA_ADDR_OFFSET, data, resize);

	//set polling flag
	flag = 1;
	write_dpram(fd, POLLING_FLAG_ADDR_OFFSET, &flag, 1);
	write_dpram(fd, PC_INT_RDC_ADDR_OFFSET, &flag, 1);

	//wait polling flag is clear
	clock_gettime(CLOCK_MONOTONIC, &start_time);
	do {
		read_dpram(fd, POLLING_FLAG_ADDR_OFFSET, &flag, 1);
		if (flag == flag_value)
			break;

		clock_nanosleep(CLOCK_REALTIME, 0, &polling_time, NULL);
		clock_gettime(CLOCK_MONOTONIC, &end_time);
		if(DIFF_NS(end_time, start_time) >= timeout_ns) {
			flag = 0;
			write_dpram(fd, POLLING_FLAG_ADDR_OFFSET, &flag, 1);
			write_dpram(fd, PC_INT_RDC_ADDR_OFFSET, &flag, 1);
			return ECAT_ERR_IPC_TIME_OUT;
		}

	} while(flag != flag_value);

	return 0;
}
/*****************************************************************************/
int32_t SetCommandData(ecat_device_ipc_t *EcatDevice, int32_t CmdNo, void *Data, uint32_t Size)
{
	return __SetCommandData(EcatDevice->Fd, CmdNo, (uint32_t *)Data, Size, 0, 0);
}
/*****************************************************************************/
int32_t SetCommandData_OpenDevice(ecat_device_ipc_t *EcatDevice, int32_t CmdNo, void *Data, uint32_t Size)
{
	return __SetCommandData(EcatDevice->Fd, CmdNo, (uint32_t *)Data, Size, 2, 0);
}
/*****************************************************************************/
int32_t SetCommandDataEx(ecat_device_ipc_t *EcatDevice, int32_t CmdNo, void *Data, uint32_t Size, uint64_t TimeOut_ms)
{
	return __SetCommandData(EcatDevice->Fd, CmdNo, (uint32_t *)Data, Size, 0, TimeOut_ms);
}
/*****************************************************************************/
int32_t __GetReturnData(int fd, uint32_t *data,uint32_t size)
{
	uint32_t resize = size / ECAT_IPC_SHM_ALIGN;
	int32_t value;

	//get return value
	read_dpram(fd, CMD_RET_ADDR_OFFSET, &value, 1);

	//get return data
	if(data != NULL)
		read_dpram(fd, DATA_ADDR_OFFSET, data, resize);

	return value;
}
///*****************************************************************************/
int32_t GetReturnData(ecat_device_ipc_t *EcatDevice, void *Data, uint32_t Size)
{
	 return __GetReturnData(EcatDevice->Fd, (uint32_t *)Data, Size);
}
/*****************************************************************************/
int32_t GenericSendCommand(uint16_t device_no, uint64_t timeout_ms, int32_t cmd_no, void *args, uint32_t args_size, void *ret_data, uint32_t ret_size)
{
	CHECK_ECAT_DEVICE_NO(device_no)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[device_no].IsOpen)
	int32_t ret;
	SEMEPHORE_P_OPERATION(EcatDevice[device_no].ThreadSem)

	if ((ret = __SetCommandData(EcatDevice[device_no].Fd, cmd_no, (uint32_t *)args, args_size, 0, timeout_ms)) < 0)
		goto out_return;

	ret = __GetReturnData(EcatDevice[device_no].Fd, (uint32_t *)ret_data, ret_size);

out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[device_no].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t irq_setup(uint16_t DeviceNo, uint8_t Enabled)
{
	ixecat_reg_t reg;
	if (Enabled) {
		if (EcatDevice[DeviceNo].irq_enable) {
			return 0;
		}
	} else {
		if (!EcatDevice[DeviceNo].irq_enable)  {
			return 0;
		}
	}

	reg.id = IXECAT_ISR;    //CPU(or Host) side's ioctl command
	reg.value = Enabled; //enable En_ISR
	if (ioctl(EcatDevice[DeviceNo].Fd, IXECAT_WRITE_REG, &reg)) {
		return ECAT_ERR_DRV_IOCTL;
	} else {
		EcatDevice[DeviceNo].irq_enable = Enabled;
		return 0;
	}
}
/*****************************************************************************/
static void sig_handler(int sig)
{
	int DeviceNo = sig - SIGRTMIN;
	uint32_t state;
	if ((DeviceNo >= CARD_DEVICE_NO_MAX) && (DeviceNo < 0))
		return;

	read_dpram(EcatDevice[DeviceNo].Fd, 0x7FE, &state, 1);
	state = (state >> 8) & 0xFF;
	if (state & 0x01)
		pthread_cond_broadcast(&EcatDevice[DeviceNo].cond_timer);

	if (state & 0x04)
		pthread_cond_broadcast(&EcatDevice[DeviceNo].cond_event);
}
/*****************************************************************************/
int32_t ECAT_GetDllVersion(char *Version, uint16_t *Size)
{
	if (Version) {
		strcpy(Version, LINUX_LIB_VERSION);
	} else {
		if (Size)
			*Size = sizeof(LINUX_LIB_VERSION);
	}
	return 0;
}
/*****************************************************************************/
int32_t ECAT_GetFirmwareVersion(uint16_t DeviceNo, char *Version, uint16_t *Size)
{
	int32_t ret;
	ecat_firmware_info_t RetData;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_FIRWARE_VERSION,
							NULL, 0, &RetData, sizeof(ecat_firmware_info_t));

	if (ret == 0) {
		if (Version)
			memcpy(Version, RetData.Data, sizeof(RetData.Data) + 1);
		else
			if (Size)
				*Size = RetData.Size;
	}
	return ret;
}
/*****************************************************************************/
int32_t __GetDeviceCnt(uint16_t *DeviceCnt)
{
	uint16_t i;
	uint16_t cnt = 0;
	char dev_file[128];
	for (i = 1; i <= CARD_DEVICE_NO_MAX; i++) {
		sprintf(dev_file, "/dev/ixecat%d", i);
		if (access( dev_file, F_OK ) != -1)
			cnt++;
	}
	if (cnt > CARD_DEVICE_NO_MAX)
		cnt = CARD_DEVICE_NO_MAX;

	*DeviceCnt = cnt;
	return 0;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceCnt(uint16_t *DeviceCnt, uint8_t CardID[])
{
	int32_t ret;
	uint16_t cnt = 0;
	uint16_t i;
	uint32_t card_id;
	int fd_dev;
	char dev_file[128];

	__GetDeviceCnt(&cnt);
	*DeviceCnt = cnt;

	for (i = 1; i <= cnt; i++) {
		sprintf(dev_file, "/dev/ixecat%d", i);
		fd_dev = open(dev_file, O_RDWR);
		if (fd_dev == -1)
			return ECAT_ERR_IPC_DEVICE_NOT_OPEN;
		else {
			if ((ret = read_bar0_reg(fd_dev, CARD_ID_ADDR_OFFSET, &card_id, 1) != 0))
				return ret;
			else {
				card_id = ~card_id & 0xF;
				CardID[i - 1] = card_id;
				close(fd_dev);
			}
		}
	}
	return 0;
}
/*****************************************************************************/
int32_t ECAT_OpenDevice(uint16_t DeviceNo)
{
	int32_t ret;
	uint16_t i;
	uint16_t cnt;
	uint32_t card_id;
	bool finde_dev = false;
	int fd_dev;
	char dev_file[128];
	uint32_t dpram_data;
	struct sigaction *sig_act = &EcatDevice[DeviceNo].sig_act;
	ixecat_signal_t sig;

	CHECK_ECAT_DEVICE_NO(DeviceNo)
	if (EcatDevice[DeviceNo].IsOpen)
		return 0;

	__GetDeviceCnt(&cnt);
	for (i = 1; i <= cnt; i++) {
		sprintf(dev_file, "/dev/ixecat%d", i);
		fd_dev = open(dev_file, O_RDWR);
		if (fd_dev == -1)
			continue;
		else {
			if ((ret = read_bar0_reg(fd_dev, CARD_ID_ADDR_OFFSET, &card_id, 1) != 0)) {
				close(fd_dev);
				continue;
			}
			else {

				card_id = ~card_id & 0xF;
				if (card_id == DeviceNo) {
					finde_dev = true;
					break;
				} else
					close(fd_dev);
			}
		}
	}

	if (!finde_dev)
		return ECAT_ERR_DRV_DEVICE_NOT_FOUND;

	sem_init(&EcatDevice[DeviceNo].ThreadSem, 0, 1);
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	EcatDevice[DeviceNo].Fd = fd_dev;
	dpram_data = 0;
	write_dpram(EcatDevice[DeviceNo].Fd, POLLING_FLAG_ADDR_OFFSET, &dpram_data, 1);

	if ((ret = SetCommandData_OpenDevice(&EcatDevice[DeviceNo], ECAT_IPC_CMD_OPEN_DEVICE, NULL, 0)) < 0) {
		goto out_return;
	}

	write_dpram(EcatDevice[DeviceNo].Fd, POLLING_FLAG_ADDR_OFFSET, &dpram_data, 1);

	/* signal action */
	sig_act->sa_handler = sig_handler;
	sigemptyset(&sig_act->sa_mask);
	sigaddset(&sig_act->sa_mask, (SIGRTMIN + DeviceNo));
	if (sigaction((SIGRTMIN + DeviceNo), sig_act, NULL)) {
		ret = ECAT_ERR_DRV_CREATE_HANDLE;
		goto out_return;
	}

	if (pthread_mutex_init(&EcatDevice[DeviceNo].cond_timer_mutex, NULL)) {
		ret = ECAT_ERR_DRV_CREATE_HANDLE;
		goto out_return;
	}

	if (pthread_cond_init(&EcatDevice[DeviceNo].cond_event, NULL)) {
		ret = ECAT_ERR_DRV_CREATE_HANDLE;
		goto out_return;
	}

	if (pthread_mutex_init(&EcatDevice[DeviceNo].cond_event_mutex, NULL)) {
		ret = ECAT_ERR_DRV_CREATE_HANDLE;
		goto out_return;
	}

	if (pthread_cond_init(&EcatDevice[DeviceNo].cond_timer, NULL)) {
		ret = ECAT_ERR_DRV_CREATE_HANDLE;
		goto out_return;
	}

	/* signal condition */
	sig.sid = (SIGRTMIN + DeviceNo);
	sig.pid = getpid();
	if (ioctl(EcatDevice[DeviceNo].Fd, IXECAT_SET_SIG, &sig)) {
		ret = ECAT_ERR_DRV_IOCTL;
		goto out_return;
	}

	for (i = 0; i < MC_GROUP_NO_MAX; i++)
		EcatDevice[DeviceNo].GrpAxisCnt[i] = 0;
	EcatDevice[DeviceNo].IsOpen = true;
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return 0;

out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	close(fd_dev);
	EcatDevice[DeviceNo].IsOpen = false;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_CloseDevice(uint16_t DeviceNo)
{
	int32_t ret;
	uint16_t i;

	CHECK_ECAT_DEVICE_NO(DeviceNo)
	if (!EcatDevice[DeviceNo].IsOpen)
		return 0;

	if ((!EcatDevice[DeviceNo].MpgThreadShouldStop) && (EcatDevice[DeviceNo].MpgThreadRunning)) {
		SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
		EcatDevice[DeviceNo].MpgThreadShouldStop = 1;
		SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
		pthread_join(EcatDevice[DeviceNo].MpgThreadID, NULL);

	}

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	ret = irq_setup(DeviceNo, 0);
	if (ret != 0) {
		goto out_return;
	}

	SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_STOP_DEV_OP_TASK, NULL, 0);

	if ((ret = close(EcatDevice[DeviceNo].Fd)) == 0)
	{
		EcatDevice[DeviceNo].IsOpen = false;
		for (i = 0; i<MC_GROUP_NO_MAX; i++)
			EcatDevice[DeviceNo].GrpAxisCnt[i] = 0;
		ret = 0;
		goto out_return;
	}
	else
	{
		ret = ECAT_ERR_DRV_IOCTL;
		goto out_return;
	}

out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceSerialNo(uint16_t DeviceNo, uint8_t *SerialNo)
{
	int32_t ret;
	ecat_read_file_return_t RetData;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_DEVICE_SERIAL_NO,
							NULL, 0, &RetData, sizeof(ecat_read_file_return_t));
	printf("new ECAT_GetDeviceSerialNo\n");
	if (ret == 0) {
		memcpy(SerialNo, RetData.Data, sizeof(uint8_t) * DEVICE_SERIAL_NO_SIZE);
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceDI(uint16_t DeviceNo, uint32_t *Value)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret;
	uint32_t DI = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, GPIO_ADDR_OFFSET, &DI, 1) != 0))
		goto out_return;
	DI = ~(DI >> 16);
	DI &= 0x1FFF;
	*Value = DI;
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceDIBit(uint16_t DeviceNo, uint16_t BitNo, uint32_t *Value)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	uint32_t BitValue = 1 << BitNo;
	int32_t ret;
	uint32_t DI = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, GPIO_ADDR_OFFSET, &DI, 1) != 0))
		goto out_return;

	DI = ~(DI >> 16);
	DI &= 0x1FFF;
	DI &= BitValue;
	if (DI)
		*Value = 1;
	else
		*Value = 0;
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceDO(uint16_t DeviceNo, uint32_t *Value)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret;
	uint32_t DO = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, GPIO_ADDR_OFFSET, &DO, 1) != 0))
		goto out_return;
	DO &= 0xFFFF;
	*Value = DO;
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceDOBit(uint16_t DeviceNo, uint16_t BitNo, uint32_t *Value)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	uint32_t BitValue = 1 << BitNo;
	int32_t ret;
	uint32_t DO = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, GPIO_ADDR_OFFSET, &DO, 1) != 0))
		goto out_return;
	DO &= BitValue;
	if (DO)
		*Value = 1;
	else
		*Value = 0;
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetDeviceDO(uint16_t DeviceNo, uint32_t Value)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret;
	uint32_t DO = Value;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	DO &= 0x1FFF;
	ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, GPIO_ADDR_OFFSET, &DO, 1);
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetDeviceDOBit(uint16_t DeviceNo, uint16_t BitNo, uint32_t Value)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	uint32_t BitValue = 1 << BitNo;
	int32_t ret;
	uint32_t DO = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, GPIO_ADDR_OFFSET, &DO, 1) != 0))
		goto out_return;
	DO &= 0xFFFF;
	if (Value)
		DO |= BitValue;
	else
		DO &= ~BitValue;
	ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, GPIO_ADDR_OFFSET, &DO, 1);
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceState(uint16_t DeviceNo, uint32_t *LinkUp, uint32_t *SlavesResp, uint32_t *AlStates, uint32_t *Wc)
{
	int32_t ret;
	ecat_device_state_t RetData;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_DEVICE_STATE,
							NULL, 0, &RetData, sizeof(ecat_device_state_t));
	if (ret == 0) {
		*LinkUp = RetData.LinkUp;
		*SlavesResp = RetData.SlavesResp;
		*AlStates = RetData.AlStates;
		*Wc = RetData.Wc;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveInfo(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t *Alias, uint32_t *ProductCode, uint32_t *VendorID,
	uint32_t *RevisionNo, uint32_t *SerialNo, uint8_t *AlState, uint32_t *SlaveType, char *Slavename)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_slave_info_t RetData;
	Ops.No = SlaveNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_SLAVE_INFO,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_slave_info_t));
	if (ret == 0) {
		*Alias = RetData.Alias;
		*ProductCode = RetData.ProductCode;
		*VendorID = RetData.VendorID;
		*RevisionNo = RetData.RevisionNo;
		*SerialNo = RetData.SerialNo;
		*AlState = RetData.AlState;
		*SlaveType = RetData.SlaveType;
		if (Slavename != NULL)
			strcpy(Slavename, RetData.Slavename);
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_UploadNetworkInfo(uint16_t DeviceNo, uint16_t NetworkInfoNo, uint32_t Offset, char *Data, uint32_t DataSize, uint8_t LastFlag)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret;
	ecat_write_file_ops_t Ops;
	ecat_write_file_retrun_t RetData;
	uint32_t OpsSize = sizeof(ecat_write_file_ops_t);
	uint32_t RetSize = sizeof(ecat_write_file_retrun_t);
	uint32_t DataOffset = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	Ops.No = NetworkInfoNo;


	while (1)
	{
		if (DataSize > ECAT_IPC_DATA_SIZE_MAX)
		{
			Ops.Size = ECAT_IPC_DATA_SIZE_MAX;
			Ops.Offset = Offset + DataOffset;
			Ops.LastFlag = 0;
		}
		else
		{
			Ops.Size = DataSize;
			Ops.Offset = Offset + DataOffset;
			if (LastFlag)
				Ops.LastFlag = 1;
			else
				Ops.LastFlag = 0;
		}

		memcpy(Ops.Data, (Data + (DataOffset)), Ops.Size);

		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_UPLOAD_NETWORK_INFO, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		if ((ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&RetData, RetSize)) < 0)
			goto out_return;
		else
		{
			DataOffset += RetData.WriteBytes;
			DataSize -= RetData.WriteBytes;
			if (DataSize == 0)
			{
				ret = 0;
				goto out_return;
			}
		}
	}
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_UploadFirmware(uint16_t DeviceNo, uint32_t Offset, char *Data, uint32_t DataSize, uint8_t LastFlag)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret;
	ecat_write_file_ops_t Ops;
	ecat_write_file_retrun_t RetData;
	uint32_t OpsSize = sizeof(ecat_write_file_ops_t);
	uint32_t RetSize = sizeof(ecat_write_file_retrun_t);
	uint32_t DataOffset = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	while (1)
	{
		if (DataSize > ECAT_IPC_DATA_SIZE_MAX)
		{
			Ops.Size = ECAT_IPC_DATA_SIZE_MAX;
			Ops.Offset = Offset + DataOffset;
			Ops.LastFlag = 0;
		}
		else
		{
			Ops.Size = DataSize;
			Ops.Offset = Offset + DataOffset;
			if (LastFlag)
				Ops.LastFlag = 1;
			else
				Ops.LastFlag = 0;
		}

		memcpy(Ops.Data, (Data + (DataOffset)), Ops.Size);

		if ((ret = SetCommandDataEx(&EcatDevice[DeviceNo], ECAT_IPC_CMD_UPLOAD_FIRMWARE, (uint32_t*)&Ops, OpsSize, 2000)) < 0)
			goto out_return;

		if ((ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&RetData, RetSize)) < 0)
			goto out_return;
		else
		{
			DataOffset += RetData.WriteBytes;
			DataSize -= RetData.WriteBytes;
			if (DataSize == 0)
			{
				ret = 0;
				goto out_return;
			}
		}
	}
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_DownloadNetworkInfo(uint16_t DeviceNo, uint16_t NetworkInfoNo, uint32_t *Offset, char *Data, uint32_t *DataSize, uint8_t *LastFlag)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret;
	ecat_read_file_ops_t Ops;
	ecat_read_file_return_t RetData;
	uint32_t OpsSize = sizeof(ecat_read_file_ops_t);
	uint32_t RetSize = sizeof(ecat_read_file_return_t);
	uint32_t DataOffset = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)

	Ops.No = NetworkInfoNo;
	if (*LastFlag == 255)
	{
		Ops.GetFileSize = 1;
		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_DOWNLOAD_NETWORK_INFO, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		if ((ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&RetData, RetSize)) < 0)
			goto out_return;
		else
		{
			*DataSize = RetData.Size;
			ret = 0;
			goto out_return;
		}
	}
	else
		Ops.GetFileSize = 0;

	while (1)
	{
		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_DOWNLOAD_NETWORK_INFO, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		if ((ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&RetData, RetSize)) < 0)
			goto out_return;
		else
		{
			memcpy((Data + (DataOffset)), RetData.Data, RetData.Size);
			DataOffset += RetData.Size;
			if ((RetData.LastFlag) || (*DataSize < (DataOffset + ECAT_IPC_DATA_SIZE_MAX)))
			{
				*DataSize = DataOffset;
				*Offset = RetData.Offset;
				*LastFlag = RetData.LastFlag;
				ret = 0;
				goto out_return;
			}
		}
	}
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveSdoObject(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t Index, uint8_t SubIndex, uint16_t DataSize,
	uint32_t *ObjectVal, uint32_t *AbortCode)
{
	int32_t ret;
	ecat_slave_sdo_ops_t Ops;
	ecat_slave_sdo_return_t RetData;
	Ops.No = SlaveNo;
	Ops.Index = Index;
	Ops.SubIndex = SubIndex;
	Ops.DataSize = DataSize;
	ret = GenericSendCommand(DeviceNo, 1000, ECAT_IPC_CMD_GET_SDO_OBJECT,
							&Ops, sizeof(ecat_slave_sdo_ops_t), &RetData, sizeof(ecat_slave_sdo_return_t));
	if (ret == 0)
		*ObjectVal = RetData.ObjectVal;
	else
		*AbortCode = RetData.AbortCode;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetSlaveSdoObject(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t Index, uint8_t SubIndex, uint16_t DataSize,
	uint32_t ObjectVal, uint32_t *AbortCode)
{
	int32_t ret;
	ecat_slave_sdo_ops_t Ops;
	ecat_slave_sdo_return_t RetData;
	Ops.No = SlaveNo;
	Ops.Index = Index;
	Ops.SubIndex = SubIndex;
	Ops.DataSize = DataSize;
	Ops.ObjectVal = ObjectVal;
	ret = GenericSendCommand(DeviceNo, 1000, ECAT_IPC_CMD_SET_SDO_OBJECT,
							&Ops, sizeof(ecat_slave_sdo_ops_t), &RetData, sizeof(ecat_slave_sdo_return_t));
	if (ret != 0)
		*AbortCode = RetData.AbortCode;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SlaveNonBlockSdoRequestRead(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t Index, uint8_t SubIndex, uint32_t Timeout_ms)
{
	ecat_slave_sdo_ops_t Ops;
	Ops.No = SlaveNo;
	Ops.Index = Index;
	Ops.SubIndex = SubIndex;
	Ops.Timeout_ms = Timeout_ms;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_NONBLOCK_SDO_REQ_READ,
							&Ops, sizeof(ecat_slave_sdo_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_SlaveNonBlockSdoRequestReadState(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t DataSize, uint32_t *ObjectVal)
{
	int32_t ret;
	ecat_slave_sdo_ops_t Ops;
	ecat_slave_sdo_return_t RetData;
	Ops.No = SlaveNo;
	Ops.DataSize = DataSize;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_NONBLOCK_SDO_REQ_READ_STATE,
							&Ops, sizeof(ecat_slave_sdo_ops_t), &RetData, sizeof(ecat_slave_sdo_return_t));
	if (ret == 0)
		*ObjectVal = RetData.ObjectVal;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SlaveNonBlockSdoRequestWrite(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t Index, uint8_t SubIndex, uint16_t DataSize, uint32_t ObjectVal, uint32_t Timeout_ms)
{
	ecat_slave_sdo_ops_t Ops;
	Ops.No = SlaveNo;
	Ops.Index = Index;
	Ops.SubIndex = SubIndex;
	Ops.DataSize = DataSize;
	Ops.ObjectVal = ObjectVal;
	Ops.Timeout_ms = Timeout_ms;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_NONBLOCK_SDO_REQ_WRITE,
							&Ops, sizeof(ecat_slave_sdo_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_SlaveNonBlockSdoRequestWriteState(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t DataSize)
{
	ecat_slave_sdo_ops_t Ops;
	Ops.No = SlaveNo;
	Ops.DataSize = DataSize;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_NONBLOCK_SDO_REQ_WRITE_STATE,
							&Ops, sizeof(ecat_slave_sdo_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_StartDeviceOpTask(uint16_t DeviceNo, uint16_t NetworkInfoNo, uint8_t EnumCycleTime, uint32_t WcErrCnt)
{
	ecat_start_op_task_ops_t Ops;
	Ops.No = NetworkInfoNo;
	Ops.EnumCycleTime = EnumCycleTime;
	Ops.WcErrCnt = WcErrCnt;
	return GenericSendCommand(DeviceNo, 2000, ECAT_IPC_CMD_START_DEV_OP_TASK,
							&Ops, sizeof(ecat_slave_sdo_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_StopDeviceOpTask(uint16_t DeviceNo)
{
	return GenericSendCommand(DeviceNo, 2000, ECAT_IPC_CMD_STOP_DEV_OP_TASK,
							NULL, 0, NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_SetSlaveRxPdoData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t OffsetByte, uint16_t DataSize, uint8_t *Data)
{
	ecat_slave_set_pdo_ops_t Ops;
	Ops.No = SlaveNo;
	Ops.OffsetByte = OffsetByte;
	Ops.DataSize = DataSize;
	memcpy(Ops.Data, Data, DataSize);
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_RXPDO_DATA,
							&Ops, sizeof(ecat_slave_set_pdo_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_GetSlaveRxPdoData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t OffsetByte, uint16_t DataSize, uint8_t *Data)
{
	if (DataSize >= RW_PDO_DATA_SIZE_MAX)
		return ECAT_ERR_INVALID_DATA_SIZE;
	int32_t ret;
	ecat_slave_get_pdo_ops_t Ops;
	ecat_slave_get_pdo_return_t RetData;
	Ops.No = SlaveNo;
	Ops.OffsetByte = OffsetByte;
	Ops.DataSize = DataSize;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_RXPDO_DATA,
							&Ops, sizeof(ecat_slave_get_pdo_ops_t), &RetData, sizeof(ecat_slave_get_pdo_return_t));
	if (ret == 0)
		memcpy(Data, RetData.Data, DataSize);
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveTxPdoData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t OffsetByte, uint16_t DataSize, uint8_t *Data)
{
	if (DataSize >= RW_PDO_DATA_SIZE_MAX)
		return ECAT_ERR_INVALID_DATA_SIZE;
	int32_t ret;
	ecat_slave_get_pdo_ops_t Ops;
	ecat_slave_get_pdo_return_t RetData;
	Ops.No = SlaveNo;
	Ops.OffsetByte = OffsetByte;
	Ops.DataSize = DataSize;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_TXPDO_DATA,
							&Ops, sizeof(ecat_slave_get_pdo_ops_t), &RetData, sizeof(ecat_slave_get_pdo_return_t));
	if (ret == 0)
		memcpy(Data, RetData.Data, DataSize);
	return ret;
}
/*****************************************************************************/
int32_t CommonSlaveDIO(uint16_t DeviceNo, uint32_t RunMode, uint16_t SlaveNo, uint16_t BitNo, uint32_t *Value)
{
	int32_t ret;
	ecat_slave_dio_data_t Ops;
	Ops.RunMode = RunMode;
	Ops.SlaveNo = SlaveNo;
	Ops.BitNo = BitNo;
	Ops.U32Value = *Value;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SLAVE_DIO,
							&Ops, sizeof(ecat_slave_dio_data_t), &Ops, sizeof(ecat_slave_dio_data_t));
	if ((RunMode != SLAVE_DIO_RM_SET_DO) && (RunMode != SLAVE_DIO_RM_SET_DO_BIT)) {
		if (ret == 0)
			*Value = Ops.U32Value;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveDI(uint16_t DeviceNo, uint16_t SlaveNo, uint32_t *Value)
{
	return CommonSlaveDIO(DeviceNo, SLAVE_DIO_RM_GET_DI, SlaveNo, 0, Value);
}
/*****************************************************************************/
int32_t ECAT_GetSlaveDIBit(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t BitNo, uint32_t *Value)
{
	return CommonSlaveDIO(DeviceNo, SLAVE_DIO_RM_GET_DI_BIT, SlaveNo, BitNo, Value);
}
/*****************************************************************************/
int32_t ECAT_GetSlaveDO(uint16_t DeviceNo, uint16_t SlaveNo, uint32_t *Value)
{
	return CommonSlaveDIO(DeviceNo, SLAVE_DIO_RM_GET_DO, SlaveNo, 0, Value);
}
/*****************************************************************************/
int32_t ECAT_GetSlaveDOBit(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t BitNo, uint32_t *Value)
{
	return CommonSlaveDIO(DeviceNo, SLAVE_DIO_RM_GET_DO_BIT, SlaveNo, BitNo, Value);
}
/*****************************************************************************/
int32_t ECAT_SetSlaveDO(uint16_t DeviceNo, uint16_t SlaveNo, uint32_t Value)
{
	return CommonSlaveDIO(DeviceNo, SLAVE_DIO_RM_SET_DO, SlaveNo, 0, &Value);
}
/*****************************************************************************/
int32_t ECAT_SetSlaveDOBit(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t BitNo, uint32_t Value)
{
	return CommonSlaveDIO(DeviceNo, SLAVE_DIO_RM_SET_DO_BIT, SlaveNo, BitNo, &Value);
}
/*****************************************************************************/
int32_t CommonSlaveAIO(uint16_t DeviceNo, ecat_slave_aio_data_t *Ops)
{
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SLAVE_AIO,
							Ops, sizeof(ecat_slave_aio_data_t), Ops, sizeof(ecat_slave_aio_data_t));
}
/*****************************************************************************/
int32_t ECAT_SetSlaveAoProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, uint8_t Range)
{
	int32_t ret;
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_SET_AO_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	Ops.Range = Range;
	ret = CommonSlaveAIO(DeviceNo, &Ops);
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveAoProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, uint8_t *Range)
{
	int32_t ret;
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_GET_AO_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	ret = CommonSlaveAIO(DeviceNo, &Ops);
	if (ret == 0)
		*Range = Ops.Range;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetSlaveAoRawData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, int16_t Data)
{
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_SET_AO_RAW_DATA;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	Ops.RawData = Data;
	return CommonSlaveAIO(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_GetSlaveAoRawData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, int16_t *Data)
{
	int32_t ret;
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_GET_AO_DATA;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	ret = CommonSlaveAIO(DeviceNo, &Ops);
	if (ret == 0)
		*Data = Ops.RawData;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetSlaveAoVoltData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, double Data)
{
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_SET_AO_VOLT_DATA;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	Ops.VoltData = Data;
	return CommonSlaveAIO(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_GetSlaveAoVoltData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, double *Data)
{
	int32_t ret;
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_GET_AO_DATA;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	ret = CommonSlaveAIO(DeviceNo, &Ops);
	if (ret == 0)
		*Data = Ops.VoltData;
	return ret;
}
/*****************************************************************************/
int32_t CommonSlaveEnc(uint16_t DeviceNo, ecat_slave_enc_data_t *Ops)
{
	return GenericSendCommand(DeviceNo, 200, ECAT_IPC_CMD_SLAVE_ENC,
							Ops, sizeof(ecat_slave_enc_data_t), Ops, sizeof(ecat_slave_enc_data_t));
}
/*****************************************************************************/
int32_t ECAT_SetSlaveEncProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, uint8_t Mode, uint8_t InvertCnt, uint8_t LPF)
{
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_SET_ENC_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	Ops.Arg[0] = Mode;
	Ops.Arg[1] = InvertCnt;
	Ops.Arg[2] = LPF;
	return CommonSlaveEnc(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_GetSlaveEncProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, uint8_t *Mode, uint8_t *InvertCnt, uint8_t *LPF)
{
	int32_t ret;
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_GET_ENC_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	ret = CommonSlaveEnc(DeviceNo, &Ops);
	if (ret == 0)
	{
		*Mode = Ops.Arg[0];
		*InvertCnt = Ops.Arg[1];
		*LPF = Ops.Arg[2];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveEncCount(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, int32_t *Cnt)
{
	int32_t ret;
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_GET_ENC_CNT;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	ret = CommonSlaveEnc(DeviceNo, &Ops);
	if (ret == 0)
		*Cnt = Ops.I32Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_ResetSlaveEncCount(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo)
{
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_RESET_ENC_CNT;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	return CommonSlaveEnc(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_SetSlaveEncCount(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, int32_t Value)
{
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_SET_ENC_CNT;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	Ops.I32Value = Value;
	return CommonSlaveEnc(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_SetSlaveEncIdxLatchProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, uint8_t Enable)
{
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_SET_ENC_IDX_LATCH_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	Ops.Arg[0] = Enable;
	return CommonSlaveEnc(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_GetSlaveEncIdxLatchProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, uint8_t *Enable)
{
	int32_t ret;
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_GET_ENC_IDX_LATCH_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	ret = CommonSlaveEnc(DeviceNo, &Ops);
	if (ret == 0) {
		*Enable = Ops.Arg[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveEncIdxLatchCnt(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, int32_t *Cnt)
{
	int32_t ret;
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_GET_ENC_IDX_LATCH_CNT;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	ret = CommonSlaveEnc(DeviceNo, &Ops);
	if (ret == 0)
		*Cnt = Ops.I32Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_ResetSlaveEncIdxLatchCnt(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo)
{
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_RESET_ENC_IDX_LATCH_CNT;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	return CommonSlaveEnc(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_SetSlaveEncExtLatchProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, uint8_t Enable, uint8_t Mode, uint8_t Logic)
{
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_SET_ENC_EXT_LATCH_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	Ops.Arg[0] = Enable;
	Ops.Arg[1] = Mode;
	Ops.Arg[2] = Logic;
	return CommonSlaveEnc(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_GetSlaveEncExtLatchProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, uint8_t *Enable, uint8_t *Mode, uint8_t *Logic)
{
	int32_t ret;
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_GET_ENC_EXT_LATCH_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	ret = CommonSlaveEnc(DeviceNo, &Ops);
	if (ret == 0) {
		*Enable = Ops.Arg[0];
		*Mode = Ops.Arg[1];
		*Logic = Ops.Arg[2];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveEncExtLatchCnt(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, int32_t *Cnt)
{
	int32_t ret;
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_GET_ENC_EXT_LATCH_CNT;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	ret = CommonSlaveEnc(DeviceNo, &Ops);
	if (ret == 0)
		*Cnt = Ops.I32Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_ResetSlaveEncExtLatchCnt(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo)
{
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_RESET_ENC_EXT_LATCH_CNT;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	return CommonSlaveEnc(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_SetSlaveCmpTrigProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, uint8_t PulseWidth)
{
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_SET_CMP_TRIG_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	Ops.Arg[0] = PulseWidth;
	return CommonSlaveEnc(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_GetSlaveCmpTrigProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, uint8_t *PulseWidth)
{
	int32_t ret;
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_GET_CMP_TRIG_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	ret = CommonSlaveEnc(DeviceNo, &Ops);
	if (ret == 0) {
		*PulseWidth = Ops.Arg[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetSlaveCmpTrigData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, int32_t CmpData)
{
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_SET_CMP_TRIG_DATA;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	Ops.I32Value = CmpData;
	return CommonSlaveEnc(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_SetSlaveContCmpTrigData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t EncNo, int32_t Start, uint32_t Interval, uint8_t Dir)
{
	ecat_slave_enc_data_t Ops;
	Ops.RunMode = SLAVE_ENC_RM_SET_CONT_CMP_TRIG_DATA;
	Ops.SlaveNo = SlaveNo;
	Ops.EncNo = EncNo;
	Ops.I32Value = Start;
	Ops.U32Value = Interval;
	Ops.Arg[0] = Dir;
	return CommonSlaveEnc(DeviceNo, &Ops);
}
/*****************************************************************************/
int32_t ECAT_McInit(uint16_t DeviceNo, uint16_t SlaveNo[], uint16_t SubAxisNo[], uint16_t AxisCount)
{
	if (AxisCount > MC_AXIS_NO_MAX)
		return ECAT_ERR_MC_AXIS_CNT_EXCEEDED;
	int32_t ret;
	ecat_mc_init_ops_t Ops;
	uint16_t i;
	memcpy(Ops.SlaveNo, SlaveNo, sizeof(uint16_t) * AxisCount);
	memcpy(Ops.SubAxisNo, SubAxisNo, sizeof(uint16_t) * AxisCount);
	Ops.AxisCount = AxisCount;
	ret =  GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_INIT,
							&Ops, sizeof(ecat_mc_init_ops_t), NULL, 0);
	if (ret == 0) 	{
		SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
		for (i = 0; i<MC_GROUP_NO_MAX; i++)
			EcatDevice[DeviceNo].GrpAxisCnt[i] = 0;
		SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisServoOn(uint16_t DeviceNo, uint16_t AxisNo, uint16_t State)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = State;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_SERVO_ON,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetAxisPPU(uint16_t DeviceNo, uint16_t AxisNo, double PPU)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = PPU;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_PPU,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisPPU(uint16_t DeviceNo, uint16_t AxisNo, double *PPU)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_PPU,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*PPU = RetData.DoubleValue[0];
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisEncoderPPR(uint16_t DeviceNo, uint16_t AxisNo, uint32_t PPR)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.U32Data = PPR;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_ENCODER_PPR,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisEncoderPPR(uint16_t DeviceNo, uint16_t AxisNo, uint32_t *PPR)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_ENCODER_PPR,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*PPR = RetData.U32Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisMotorPPR(uint16_t DeviceNo, uint16_t AxisNo, uint32_t PPR)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.U32Data = PPR;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_MOTOR_PPR,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisMotorPPR(uint16_t DeviceNo, uint16_t AxisNo, uint32_t *PPR)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_MOTOR_PPR,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*PPR = RetData.U32Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisHomeMethod(uint16_t DeviceNo, uint16_t AxisNo, int32_t Method)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.I32Data = Method;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_HOME_METHOD,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisHomeMethod(uint16_t DeviceNo, uint16_t AxisNo, int32_t *Method)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_HOME_METHOD,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*Method = RetData.I32Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisHomeSpeed(uint16_t DeviceNo, uint16_t AxisNo, double SeachSw, double SeachZr)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = SeachSw;
	Ops.DoubleData[1] = SeachZr;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_HOME_SPEED,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisHomeSpeed(uint16_t DeviceNo, uint16_t AxisNo, double *SeachSw, double *SeachZr)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_HOME_SPEED,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0) {
		*SeachSw = RetData.DoubleValue[0];
		*SeachZr = RetData.DoubleValue[1];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisHomeAcc(uint16_t DeviceNo, uint16_t AxisNo, double Acc)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Acc;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_HOME_ACC,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisHomeAcc(uint16_t DeviceNo, uint16_t AxisNo, double *Acc)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_HOME_ACC,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*Acc = RetData.DoubleValue[0];
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisHomeOffset(uint16_t DeviceNo, uint16_t AxisNo, double Offset)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Offset;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_HOME_OFFSET,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisHomeOffset(uint16_t DeviceNo, uint16_t AxisNo, double *Offset)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_HOME_OFFSET,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*Offset = RetData.DoubleValue[0];
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisAccTime(uint16_t DeviceNo, uint16_t AxisNo, uint16_t Time_ms)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = Time_ms;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_ACC_TIME,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisAccTime(uint16_t DeviceNo, uint16_t AxisNo, uint16_t *Time_ms)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_ACC_TIME,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*Time_ms = RetData.U16Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisAccDecType(uint16_t DeviceNo, uint16_t AxisNo, uint16_t Type)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = Type;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_ACC_DEC_TYPE,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisAccDecType(uint16_t DeviceNo, uint16_t AxisNo, uint16_t *Type)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_ACC_DEC_TYPE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*Type = RetData.U16Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetVirtualAxisEncoder(uint16_t DeviceNo, uint16_t AxisNo, uint16_t SlaveNo, uint16_t EncNo, uint8_t Enable)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = SlaveNo;
	Ops.U16Data[1] = EncNo;
	Ops.U32Data = Enable;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_VIRTUAL_AXIS_ENCODER,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetVirtualAxisEncoder(uint16_t DeviceNo, uint16_t AxisNo, uint16_t *SlaveNo, uint16_t *EncNo, uint8_t *Enable)
{
	int32_t ret;
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_VIRTUAL_AXIS_ENCODER,
							&Ops, sizeof(ecat_mc_axis_ops_t), &Ops, sizeof(ecat_mc_axis_ops_t));
	if (ret == 0) {
		*SlaveNo = Ops.U16Data[0];
		*EncNo = Ops.U16Data[1];
		*Enable = Ops.U32Data;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisTouchProbeProperty(uint16_t DeviceNo, uint16_t AxisNo, uint16_t ProbeNo, uint8_t Enable, uint8_t Logic)
{
	ecat_mc_axis_touch_probe_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.ProbeNo = ProbeNo;
	Ops.Enable = Enable;
	Ops.Logic = Logic;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_TOUCH_PROBE,
							&Ops, sizeof(ecat_mc_axis_touch_probe_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisTouchProbeProperty(uint16_t DeviceNo, uint16_t AxisNo, uint16_t ProbeNo, uint8_t *Enable, uint8_t *Logic)
{
	int32_t ret;
	ecat_mc_axis_touch_probe_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.ProbeNo = ProbeNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_TOUCH_PROBE,
							&Ops, sizeof(ecat_mc_axis_touch_probe_ops_t), &Ops, sizeof(ecat_mc_axis_touch_probe_ops_t));
	if (ret == 0) {
		*Enable = Ops.Enable;
		*Logic = Ops.Logic;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisTouchProbeValue(uint16_t DeviceNo, uint16_t AxisNo, uint16_t ProbeNo, double *Value)
{
	int32_t ret;
	ecat_mc_axis_touch_probe_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.ProbeNo = ProbeNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_TOUCH_PROBE_VALUE,
							&Ops, sizeof(ecat_mc_axis_touch_probe_ops_t), &Ops, sizeof(ecat_mc_axis_touch_probe_ops_t));
	if (ret == 0) {
		*Value = Ops.Value;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisState(uint16_t DeviceNo, uint16_t AxisNo, uint32_t *State)
{
	int32_t ret;
	uint16_t i;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_tt RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_STATE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_tt));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++)
				*(State + i) = RetData.U32Value32[i];
		} else
			*State = RetData.U32Value32[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisMotionState(uint16_t DeviceNo, uint16_t AxisNo, uint32_t *State)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_tt RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_MOTION_STATE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_tt));
	if (ret == 0)
		*State = RetData.U32Value32[0];
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisLastError(uint16_t DeviceNo, uint16_t AxisNo, int32_t *Error)
{
	int32_t ret;
	uint16_t i;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_tt RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_LAST_ERROR,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_tt));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++)
				*(Error + i) = RetData.I32Value32[i];
		} else
			*Error = RetData.I32Value32[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisDriveError(uint16_t DeviceNo, uint16_t AxisNo, uint16_t *Error)
{
	int32_t ret;
	uint16_t i;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_tt RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_DRIVE_ERROR,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_tt));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++)
				*(Error + i) = RetData.U16Value32[i];
		} else
			*Error = RetData.U16Value32[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisActualPos(uint16_t DeviceNo, uint16_t AxisNo, double *Pos)
{
	int32_t ret;
	uint16_t i;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_double_multi_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_ACT_POS,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_double_multi_t));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++)
				*(Pos + i) = RetData.DoubleValue_multi[i];
		} else
			*Pos = RetData.DoubleValue_multi[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisCommandPos(uint16_t DeviceNo, uint16_t AxisNo, double *Pos)
{
	int32_t ret;
	uint16_t i;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_double_multi_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_CMD_POS,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_double_multi_t));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++)
				*(Pos + i) = RetData.DoubleValue_multi[i];
		} else
			*Pos = RetData.DoubleValue_multi[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisActualVel(uint16_t DeviceNo, uint16_t AxisNo, double *Vel)
{
	int32_t ret;
	uint16_t i;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_double_multi_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_ACT_VEL,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_double_multi_t));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++)
				*(Vel + i) = RetData.DoubleValue_multi[i];
		} else
			*Vel = RetData.DoubleValue_multi[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisDI(uint16_t DeviceNo, uint16_t AxisNo, uint32_t *DI)
{
	int32_t ret;
	uint16_t i;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_tt RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_DI,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_tt));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++)
				*(DI + i) = RetData.U32Value32[i];
		} else
			*DI = RetData.U32Value32[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetProfileData(uint16_t DeviceNo, uint16_t ProfileNo, double *Data, uint16_t DataSize)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	CHECK_MC_PROFILE_NO(ProfileNo)
	if (DataSize > MC_PROFILE_DATA_MAX)
		return ECAT_ERR_INVALID_DATA_SIZE;
	int32_t ret;
	ecat_mc_rw_profile_data_ops_t Ops;
	uint32_t OpsSize = sizeof(ecat_mc_rw_profile_data_ops_t);

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	Ops.ProfileNo = ProfileNo;

	for (Ops.Offset = 0; Ops.Offset<DataSize; Ops.Offset += ECAT_IPC_PROFILE_DATA_SIZE_MAX)
	{
		Ops.Size = MIN(ECAT_IPC_PROFILE_DATA_SIZE_MAX, DataSize - Ops.Offset);
		memcpy(Ops.Data, (Data + Ops.Offset), sizeof(double)*Ops.Size);
		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_MC_SET_PROFILE_DATA, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		ret = GetReturnData(&EcatDevice[DeviceNo], NULL, 0);
		if (ret != 0)
			break;
	}
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetProfileData(uint16_t DeviceNo, uint16_t ProfileNo, double *Data, uint16_t DataSize)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	CHECK_MC_PROFILE_NO(ProfileNo)
	if (DataSize > MC_PROFILE_DATA_MAX)
		return ECAT_ERR_INVALID_DATA_SIZE;
	int32_t ret;
	ecat_mc_rw_profile_data_ops_t Ops;
	uint32_t OpsSize = sizeof(ecat_mc_rw_profile_data_ops_t);

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	Ops.ProfileNo = ProfileNo;

	for (Ops.Offset = 0; Ops.Offset<DataSize; Ops.Offset += ECAT_IPC_PROFILE_DATA_SIZE_MAX)
	{
		Ops.Size = MIN(ECAT_IPC_PROFILE_DATA_SIZE_MAX, DataSize - Ops.Offset);
		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_MC_GET_PROFILE_DATA, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&Ops, OpsSize);

		if (ret != 0)
			break;
		else
			memcpy((Data + Ops.Offset), Ops.Data, sizeof(double)*Ops.Size);

	}
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetProfileCSV(uint16_t DeviceNo, uint16_t ProfileNo, uint32_t Offset, char *Data, uint32_t DataSize, uint8_t LastFlag)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret;
	ecat_write_file_ops_t Ops;
	ecat_write_file_retrun_t RetData;
	uint32_t OpsSize = sizeof(ecat_write_file_ops_t);
	uint32_t RetSize = sizeof(ecat_write_file_retrun_t);
	uint32_t DataOffset = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	Ops.No = ProfileNo;


	while (1)
	{
		if (DataSize > ECAT_IPC_DATA_SIZE_MAX)
		{
			Ops.Size = ECAT_IPC_DATA_SIZE_MAX;
			Ops.Offset = Offset + DataOffset;
			Ops.LastFlag = 0;
		}
		else
		{
			Ops.Size = DataSize;
			Ops.Offset = Offset + DataOffset;
			if (LastFlag)
				Ops.LastFlag = 1;
			else
				Ops.LastFlag = 0;
		}

		memcpy(Ops.Data, (Data + (DataOffset)), Ops.Size);

		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_MC_SET_PROFILE_CSV, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		if ((ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&RetData, RetSize)) < 0)
			goto out_return;
		else
		{
			DataOffset += RetData.WriteBytes;
			DataSize -= RetData.WriteBytes;
			if (DataSize == 0)
			{
				ret = 0;
				goto out_return;
			}
		}
	}
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetProfileCSV(uint16_t DeviceNo, uint16_t ProfileNo, uint32_t *Offset, char *Data, uint32_t *DataSize, uint8_t *LastFlag)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret;
	ecat_read_file_ops_t Ops;
	ecat_read_file_return_t RetData;
	uint32_t OpsSize = sizeof(ecat_read_file_ops_t);
	uint32_t RetSize = sizeof(ecat_read_file_return_t);
	uint32_t DataOffset = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)

	Ops.No = ProfileNo;
	if (*LastFlag == 255)
	{
		Ops.GetFileSize = 1;
		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_MC_GET_PROFILE_CSV, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		if ((ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&RetData, RetSize)) < 0)
			goto out_return;
		else
		{
			*DataSize = RetData.Size;
			ret = 0;
			goto out_return;
		}
	}
	else
		Ops.GetFileSize = 0;

	while (1)
	{
		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_MC_GET_PROFILE_CSV, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		if ((ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&RetData, RetSize)) < 0)
			goto out_return;
		else
		{
			memcpy((Data + (DataOffset)), RetData.Data, RetData.Size);
			DataOffset += RetData.Size;
			if ((RetData.LastFlag) || (*DataSize < (DataOffset + ECAT_IPC_DATA_SIZE_MAX)))
			{
				*DataSize = DataOffset;
				*Offset = RetData.Offset;
				*LastFlag = RetData.LastFlag;
				ret = 0;
				goto out_return;
			}
		}
	}
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetEcamTable(uint16_t DeviceNo, uint16_t TableNo, double *Data, uint16_t DataSize)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	CHECK_MC_ECAM_TABLE_NO(TableNo)
	if (DataSize > MC_ECAM_TABLE_DATA_MAX)
		return ECAT_ERR_INVALID_DATA_SIZE;
	int32_t ret;
	ecat_mc_rw_double_data_ops_t Ops;
	uint32_t OpsSize = sizeof(ecat_mc_rw_double_data_ops_t);

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	Ops.No = TableNo;

	for (Ops.Offset = 0; Ops.Offset < DataSize; Ops.Offset += ECAT_IPC_RW_DOUBLE_DATA_SIZE_MAX)
	{
		Ops.Size = MIN(ECAT_IPC_RW_DOUBLE_DATA_SIZE_MAX, DataSize - Ops.Offset);
		memcpy(Ops.Data, (Data + Ops.Offset), sizeof(double)*Ops.Size);
		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_MC_SET_ECAM_TABLE, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		ret = GetReturnData(&EcatDevice[DeviceNo], NULL, 0);
		if (ret != 0)
			break;
	}
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetEcamTable(uint16_t DeviceNo, uint16_t TableNo, double *Data, uint16_t DataSize)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	CHECK_MC_ECAM_TABLE_NO(TableNo)
	if (DataSize > MC_ECAM_TABLE_DATA_MAX)
		return ECAT_ERR_INVALID_DATA_SIZE;
	int32_t ret;
	ecat_mc_rw_double_data_ops_t Ops;
	uint32_t OpsSize = sizeof(ecat_mc_rw_double_data_ops_t);

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
		Ops.No = TableNo;

	for (Ops.Offset = 0; Ops.Offset < DataSize; Ops.Offset += ECAT_IPC_RW_DOUBLE_DATA_SIZE_MAX)
	{
		Ops.Size = MIN(ECAT_IPC_RW_DOUBLE_DATA_SIZE_MAX, DataSize - Ops.Offset);
		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_MC_GET_ECAM_TABLE, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&Ops, OpsSize);

		if (ret != 0)
			break;
		else
			memcpy((Data + Ops.Offset), Ops.Data, sizeof(double)*Ops.Size);
	}
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McConfigEcamTable(uint16_t DeviceNo, uint16_t TableNo, uint8_t SlaveAbs)
{
	ecat_mc_config_ecam_ops_t Ops;
	Ops.No = TableNo;
	Ops.SlaveAbs = SlaveAbs;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_CONFIG_ECAM_TABLE,
							&Ops, sizeof(ecat_mc_config_ecam_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisHome(uint16_t DeviceNo, uint16_t AxisNo)
{
	ecat_ipc_base_ops_t Ops;
	Ops.No = AxisNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_HOME,
							&Ops, sizeof(ecat_ipc_base_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisErrorReset(uint16_t DeviceNo, uint16_t AxisNo)
{
	ecat_ipc_base_ops_t Ops;
	Ops.No = AxisNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_ERROR_RESET,
							&Ops, sizeof(ecat_ipc_base_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisStop(uint16_t DeviceNo, uint16_t AxisNo)
{
	ecat_ipc_base_ops_t Ops;
	Ops.No = AxisNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_STOP,
							&Ops, sizeof(ecat_ipc_base_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisQuickStop(uint16_t DeviceNo, uint16_t AxisNo)
{
	ecat_ipc_base_ops_t Ops;
	Ops.No = AxisNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_QUICK_STOP,
							&Ops, sizeof(ecat_ipc_base_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveAbs(uint16_t DeviceNo, uint16_t AxisNo, double Pos, double Vel)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Pos;
	Ops.Vel = Vel;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_MOVE_ABS,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveRel(uint16_t DeviceNo, uint16_t AxisNo, double Pos, double Vel)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Pos;
	Ops.Vel = Vel;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_MOVE_REL,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisChangePos(uint16_t DeviceNo, uint16_t AxisNo, double Pos)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Pos;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_CHANGE_POS,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisChangeVel(uint16_t DeviceNo, uint16_t AxisNo, double Vel)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.Vel = Vel;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_CHANGE_VEL,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveVel(uint16_t DeviceNo, uint16_t AxisNo, double Vel)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.Vel = Vel;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_MOVE_VEL,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveVelByPos(uint16_t DeviceNo, uint16_t AxisNo, double Vel)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.Vel = Vel;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_MOVE_VEL_BY_POS,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisGearIn(uint16_t DeviceNo, uint16_t MasterNo, uint16_t SlaveNo, int32_t RatioNum, uint32_t RationDen, uint16_t SyncSource)
{
	ecat_mc_axis_ops_t Ops;
	Ops.MasterNo = MasterNo;
	Ops.AxisNo = SlaveNo;
	Ops.I32Data = RatioNum;
	Ops.U32Data = RationDen;
	Ops.U16Data[0] = SyncSource;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_GEAR_IN,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisGearOut(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t Stop)
{
	ecat_ipc_base_ops_t Ops;
	Ops.No = SlaveNo;
	switch (Stop)
	{
	case 1:
		return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_STOP,
								&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
	case 2:
		return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_QUICK_STOP,
								&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
	default:
		return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_GEAR_OUT,
								&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
	}
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveProfile(uint16_t DeviceNo, uint16_t AxisNo, uint16_t ProfileNo, uint16_t TotalStep)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = ProfileNo;
	Ops.U16Data[1] = TotalStep;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_MOVE_PROFILE,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveProfileCSV(uint16_t DeviceNo, uint16_t AxisNo, uint16_t ProfileNo)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = ProfileNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_MOVE_PROFILE_CSV,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisCamIn(uint16_t DeviceNo, uint16_t MasterNo, uint16_t SlaveNo, uint16_t TableNo, uint16_t SyncSource,
	double MasterInterval, double SlaveScaling)
{
	ecat_mc_axis_ops_t Ops;
	Ops.MasterNo = MasterNo;
	Ops.AxisNo = SlaveNo;
	Ops.U16Data[0] = TableNo;
	Ops.U16Data[1] = SyncSource;
	Ops.DoubleData[0] = MasterInterval;
	Ops.DoubleData[1] = SlaveScaling;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_CAM_IN,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisCamOut(uint16_t DeviceNo, uint16_t SlaveNo)
{
	ecat_ipc_base_ops_t Ops;
	Ops.No = SlaveNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_CAM_OUT,
							&Ops, sizeof(ecat_ipc_base_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisCamPhaseShift(uint16_t DeviceNo, uint16_t SlaveNo, double PhaseShift)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = SlaveNo;
	Ops.DoubleData[0] = PhaseShift;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_CAM_PHASE_SHIFT,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisTangentInGroup(uint16_t DeviceNo, uint16_t AxisNo, uint16_t GroupNo, double Angle, double Vel)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = GroupNo;
	Ops.DoubleData[0] = Angle;
	Ops.Vel = Vel;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_TANGENT_IN_GROUP,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAddAxisToGroup(uint16_t DeviceNo, uint16_t GroupNo, uint16_t AxisNo)
{
	int32_t ret;
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.AxisNo = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_ADD_AXIS,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
	if (ret == 0) {
		SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
		EcatDevice[DeviceNo].GrpAxisCnt[GroupNo]++;
		SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McRemoveAxisFromGroup(uint16_t DeviceNo, uint16_t GroupNo, uint16_t AxisNo)
{
	int32_t ret;
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.AxisNo = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_REMOVE_AXIS,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
	if (ret == 0) {
		SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
		EcatDevice[DeviceNo].GrpAxisCnt[GroupNo]--;
		SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McUngroupAllAxes(uint16_t DeviceNo, uint16_t GroupNo)
{
	int32_t ret;
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_UNGROUP,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
	if (ret == 0) {
		SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
		EcatDevice[DeviceNo].GrpAxisCnt[GroupNo] = 0;
		SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetGroupCmdMode(uint16_t DeviceNo, uint16_t GroupNo, uint16_t CmdMode)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = CmdMode;
	Ops.U16Data[1] = 0; //change mode immediately
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_CMD_MODE,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetGroupCmdMode(uint16_t DeviceNo, uint16_t GroupNo, uint16_t *CmdMode)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = GroupNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_GROUP_CMD_MODE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*CmdMode = RetData.U16Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetGroupState(uint16_t DeviceNo, uint16_t GroupNo, uint32_t *State)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = GroupNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_GROUP_STATE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*State = RetData.U32Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetGroupCmdBuffer(uint16_t DeviceNo, uint16_t GroupNo, uint16_t *Buffer)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = GroupNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_GROUP_CMD_BUFFER,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*Buffer = RetData.U16Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetGroupAccTime(uint16_t DeviceNo, uint16_t GroupNo, uint16_t Time_ms)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = Time_ms;
	Ops.U16Data[1] = 0; //change mode immediately
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_ACC_TIME,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetGroupAccTime(uint16_t DeviceNo, uint16_t GroupNo, uint16_t *Time_ms)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = GroupNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_GROUP_ACC_TIME,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*Time_ms = RetData.U16Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetGroupAccDecType(uint16_t DeviceNo, uint16_t GroupNo, uint16_t Type)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = Type;
	Ops.U16Data[1] = 0; //change mode immediately
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_ACC_DEC_TYPE,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetGroupAccDecType(uint16_t DeviceNo, uint16_t GroupNo, uint16_t *Type)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = GroupNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_GROUP_ACC_DEC_TYPE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
		*Type = RetData.U16Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGroupStop(uint16_t DeviceNo, uint16_t GroupNo)
{
	ecat_ipc_base_ops_t Ops;
	Ops.No = GroupNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_STOP,
							&Ops, sizeof(ecat_ipc_base_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupQuickStop(uint16_t DeviceNo, uint16_t GroupNo)
{
	ecat_ipc_base_ops_t Ops;
	Ops.No = GroupNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_QUICK_STOP,
							&Ops, sizeof(ecat_ipc_base_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t SendMcGroupMoveLine(uint16_t DeviceNo, uint16_t GroupNo, uint8_t Mode, bool AbsMove, double Vel, double Time,
	double Pos[], double pvt_vel[])
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.Vel = Vel;
	Ops.Time = Time;
	Ops.AbsMove = AbsMove;
	Ops.Mode = Mode;
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	memcpy(Ops.DoubleData, Pos, sizeof(double)*EcatDevice[DeviceNo].GrpAxisCnt[GroupNo]);
	if (Mode == MC_LINE_MODE_PVT)
		memcpy(Ops.pvt_vel, pvt_vel, sizeof(double)*EcatDevice[DeviceNo].GrpAxisCnt[GroupNo]);
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_MOVE_LINE,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveLineAbs(uint16_t DeviceNo, uint16_t GroupNo, double Pos[], double Vel)
{
	return SendMcGroupMoveLine(DeviceNo, GroupNo, MC_LINE_MODE_PV, true, Vel, 0, Pos, NULL);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveLineAbs_PT(uint16_t DeviceNo, uint16_t GroupNo, double Pos[], double Time)
{
	return SendMcGroupMoveLine(DeviceNo, GroupNo, MC_LINE_MODE_PT, true, 0, Time, Pos, NULL);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveLineAbs_PVT(uint16_t DeviceNo, uint16_t GroupNo, double Pos[], double Vel[], double Time)
{
	return SendMcGroupMoveLine(DeviceNo, GroupNo, MC_LINE_MODE_PVT, true, 0, Time, Pos, Vel);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveLineRel(uint16_t DeviceNo, uint16_t GroupNo, double Pos[], double Vel)
{
	return SendMcGroupMoveLine(DeviceNo, GroupNo, MC_LINE_MODE_PV, false, Vel, 0, Pos, NULL);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveLineRel_PT(uint16_t DeviceNo, uint16_t GroupNo, double Pos[], double Time)
{
	return SendMcGroupMoveLine(DeviceNo, GroupNo, MC_LINE_MODE_PT, false, 0, Time, Pos, NULL);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveLineRel_PVT(uint16_t DeviceNo, uint16_t GroupNo, double Pos[], double Vel[], double Time)
{
	return SendMcGroupMoveLine(DeviceNo, GroupNo, MC_LINE_MODE_PVT, false, 0, Time, Pos, Vel);
}
/*****************************************************************************/
void Vector_Cross(uint16_t Dim, double *OutVector, double *Vect1, double *Vect2)
{
	uint16_t i, j = 1, k = 2;

	for (i = 0; i<Dim; i++)
	{
		OutVector[i] = Vect1[j] * Vect2[k] - Vect1[k] * Vect2[j];
		j++; k++;
		if (j == Dim) j = 0;
		if (k == Dim) k = 0;
	}
}
/*****************************************************************************/
double Vector_Dot(uint16_t Dim, double *Vect1, double *Vect2)
{
	uint16_t i;
	double Value = 0.0;

	for (i = 0; i<Dim; i++)
		Value += Vect1[i] * Vect2[i];
	return Value;
}
/*****************************************************************************/
int32_t FindNV(double *SP, double *CP, double *EP, double *NV)
{
	int i;
	double v1[3], v2[3];
	double MNV = 0;
	for (i = 0; i < 3; i++)
	{
		v1[i] = SP[i] - CP[i];
		v2[i] = EP[i] - CP[i];
	}

	Vector_Cross(3, NV, v1, v2);

	for (i = 0; i < 3; i++)
		MNV += (NV[i] * NV[i]);

	if (MNV == 0.0)
		return -1;

	MNV = sqrt(MNV);
	for (i = 0; i < 3; i++)
		NV[i] = NV[i] / MNV;

	return 0;
}
/*****************************************************************************/
int32_t CheckNormalVector(double *Vec)
{
	int i;
	double MVec = 0;
	for (i = 0; i < 3; i++)
		MVec += (Vec[i] * Vec[i]);

	if (MVec == 0.0)
		return -1;

	MVec = sqrt(MVec);
	for (i = 0; i < 3; i++)
		Vec[i] = Vec[i] / MVec;

	return 0;
}
/*****************************************************************************/
int32_t FindCP_2D(double *p1, double *p2, double *p3, double *CP)
{
	double DeltaX21, DeltaY21, DeltaX32, DeltaY32;
	double MX1, MY1, MX2, MY2, K1, K2;
	bool	K1zero, K2zero, K1inf, K2inf;

	DeltaX21 = p2[0] - p1[0];	//(X2-X1)
	DeltaY21 = p2[1] - p1[1];	//(Y2-Y1)
	DeltaX32 = p3[0] - p2[0];	//(X3-X2)
	DeltaY32 = p3[1] - p2[1];	//(Y3-Y2)

	MX1 = (p1[0] + p2[0]) / 2;	//(X1+X2) / 2
	MY1 = (p1[1] + p2[1]) / 2;	//(Y1+Y2) / 2
	MX2 = (p2[0] + p3[0]) / 2;	//(X2+X3) / 2
	MY2 = (p2[1] + p3[1]) / 2;	//(Y2+Y3) / 2


	K1zero = (DeltaX21 <= MC_DOUBLE_EQUAL_ZERO) && (DeltaX21 >= -MC_DOUBLE_EQUAL_ZERO);
	K1inf = (DeltaY21 <= MC_DOUBLE_EQUAL_ZERO) && (DeltaY21 >= -MC_DOUBLE_EQUAL_ZERO);
	K2zero = (DeltaX32 <= MC_DOUBLE_EQUAL_ZERO) && (DeltaX32 >= -MC_DOUBLE_EQUAL_ZERO);
	K2inf = (DeltaY32 <= MC_DOUBLE_EQUAL_ZERO) && (DeltaY32 >= -MC_DOUBLE_EQUAL_ZERO);

	if ((K1zero && K2zero) || (K1inf && K2inf))
		return -1;
	else if (K1zero || K2zero)
	{
		if (K1zero)
		{
			K2 = -DeltaX32 / DeltaY32;
			CP[1] = MY1;
			CP[0] = ((CP[1] - MY2) / K2) + MX2;
		}
		else
		{
			K1 = -DeltaX21 / DeltaY21;
			CP[1] = MY2;
			CP[0] = ((CP[1] - MY1) / K1) + MX1;
		}
	}
	else if (K1inf || K2inf)
	{
		if (K1inf)
		{
			K2 = -DeltaX32 / DeltaY32;
			CP[0] = MX1;
			CP[1] = (K2 * (CP[0] - MX2)) + MY2;
		}
		else
		{
			K1 = -DeltaX21 / DeltaY21;
			CP[0] = MX2;
			CP[1] = (K1 * (CP[0] - MX1)) + MY1;
		}
	}
	else
	{
		double K1MX1, K2MX2;
		K1 = -DeltaX21 / DeltaY21;
		K2 = -DeltaX32 / DeltaY32;
		K1MX1 = K1 * MX1;
		K2MX2 = K2 * MX2;
		CP[0] = (K1MX1 - K2MX2 - MY1 + MY2) / (K1 - K2);
		CP[1] = (K1 * (CP[0] - MX1)) + MY1;
	}
	return 0;
}
/*****************************************************************************/
int32_t FindCP_3D(double *p1, double *p2, double *p3, double *CP, double *NV)
{
	double x1, y1, z1;
	double x2, y2, z2;
	double x3, y3, z3;
	double a1, b1, c1, d1;
	double a2, b2, c2, d2;
	double a3, b3, c3, d3;
	double A[3][3];
	double MNV = 0;
	int i;

	x1 = p1[0];
	y1 = p1[1];
	z1 = p1[2];
	x2 = p2[0];
	y2 = p2[1];
	z2 = p2[2];
	x3 = p3[0];
	y3 = p3[1];
	z3 = p3[2];


	// plane 1 ()
	a1 = (x2 - x1);
	b1 = (y2 - y1);
	c1 = (z2 - z1);
	d1 = (x2*x2 - x1*x1 + y2*y2 - y1*y1 + z2*z2 - z1*z1) / 2;
	// plane 2
	a2 = (x3 - x2);
	b2 = (y3 - y2);
	c2 = (z3 - z2);
	d2 = (x3*x3 - x2*x2 + y3*y3 - y2*y2 + z3*z3 - z2*z2) / 2;

	// plane 3
	a3 = (b1*c2 - b2*c1);  //cross product
	b3 = (c1*a2 - c2*a1);
	c3 = (a1*b2 - a2*b1);
	d3 = a3*x1 + b3*y1 + c3*z1;

	A[0][0] = a1;    	A[0][1] = b1;   	A[0][2] = c1;
	A[1][0] = a2;    	A[1][1] = b2;   	A[1][2] = c2;
	A[2][0] = a3;    	A[2][1] = b3;   	A[2][2] = c3;

	if (InverseMatrix((double *)A) < 0)
		return -1; // matrix inversion error!


				   /// cal center point
	CP[0] = A[0][0] * d1 + A[0][1] * d2 + A[0][2] * d3;
	CP[1] = A[1][0] * d1 + A[1][1] * d2 + A[1][2] * d3;
	CP[2] = A[2][0] * d1 + A[2][1] * d2 + A[2][2] * d3;

	NV[0] = a3;
	NV[1] = b3;
	NV[2] = c3;
	for (i = 0; i < 3; i++)
		MNV += (NV[i] * NV[i]);
	MNV = sqrt(MNV);
	for (i = 0; i < 3; i++)
		NV[i] = NV[i] / MNV;

	return 0;
}
/*****************************************************************************/
int32_t InverseMatrix(double *a)
{
	int16_t n = 3;
	int16_t is[3], js[3], i, j, k, l, u, v; /// limited usage for 3x3 matrix
	double d, p;

	for (k = 0; k <= n - 1; k++)
	{
		d = 0.0;
		for (i = k; i <= n - 1; i++)
			for (j = k; j <= n - 1; j++)
			{
				l = i*n + j;
				p = fabs(a[l]);
				if (p>d)
				{
					d = p;
					is[k] = i;
					js[k] = j;
				}
			}
		if (d + 1.0 == 1.0)
			return -1;

		if (is[k] != k)
		{
			for (j = 0; j <= n - 1; j++)
			{
				u = k*n + j;
				v = is[k] * n + j;
				p = a[u];
				a[u] = a[v];
				a[v] = p;
			}
		}
		if (js[k] != k)
		{
			for (i = 0; i <= n - 1; i++)
			{
				u = i*n + k;
				v = i*n + js[k];
				p = a[u];
				a[u] = a[v];
				a[v] = p;
			}
		}
		l = k*n + k;
		a[l] = 1.0 / a[l];
		for (j = 0; j <= n - 1; j++)
		{
			if (j != k)
			{
				u = k*n + j;
				a[u] = a[u] * a[l];
			}
		}
		for (i = 0; i <= n - 1; i++)
		{
			if (i != k)
			{
				for (j = 0; j <= n - 1; j++)
				{
					if (j != k)
					{
						u = i*n + j;
						a[u] = a[u] - a[i*n + k] * a[k*n + j];
					}
				}
			}
		}
		for (i = 0; i <= n - 1; i++)
		{
			if (i != k)
			{
				u = i*n + k;
				a[u] = -a[u] * a[l];
			}
		}

	}

	for (k = n - 1; k >= 0; k--)
	{
		if (js[k] != k)
		{
			for (j = 0; j <= n - 1; j++)
			{
				u = k*n + j;
				v = js[k] * n + j;
				p = a[u];
				a[u] = a[v];
				a[v] = p;
			}
		}
		if (is[k] != k)
		{
			for (i = 0; i <= n - 1; i++)
			{
				u = i*n + k;
				v = i*n + is[k];
				p = a[u];
				a[u] = a[v];
				a[v] = p;
			}
		}
	}
	return 0;
}
/*****************************************************************************/
int32_t SendMcGroupMoveCircHelix(uint16_t DeviceNo, uint16_t GroupNo, uint8_t CircMode, uint8_t CircDim, bool AbsMove, uint8_t MoveHelixType,
	double Vel, double Angle, uint8_t Dir, double Pitch, double AuxPos[], double EndPos[], double NV[], double EndRadius)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret = 0;
	uint16_t i;
	ecat_mc_group_ops_t Ops;
	ecat_ipc_base_ops_t Ops2;
	ecat_mc_group_ops_t RetData;
	double R, EP_Ridus;
	double x_UnitVector[3] = { 0,0,0 }, y_UnitVector[3] = { 0,0,0 };  //x' y' UnitVector, Shift rotating center to c ,
																	  //where c is the center of the circle, translate to the new coordinate system o' x' y' z',
	double x_PjtValue, y_PjtValue;  				 //x' y' projection value
	double SP_ABS[3] = { 0,0,0 };
	double CP_ABS[3] = { 0,0,0 };
	double EP_ABS[3] = { 0,0,0 };
	double SP_Rel_CP[3] = { 0,0,0 };
	double EP_Rel_CP[3] = { 0,0,0 };


	if (CircMode != MC_CIRC_MODE_CP_ANGLE) {
		Ops2.No = GroupNo;
		ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_GROUP_AXIS_CMD_POS,
								&Ops2, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_group_ops_t));
		if (ret != 0)
			goto out_return;
		else
			memcpy(SP_ABS, RetData.DoubleData, sizeof(SP_ABS));
	}

	//Calculate radius
	R = 0;
	EP_Ridus = 0;
	if (CircMode == MC_CIRC_MODE_BP_EP)
	{

		for (i = 0; i < CircDim; i++)
		{
			if (!AbsMove)
			{
				AuxPos[i] += SP_ABS[i];
				EndPos[i] += SP_ABS[i];
			}
		}


		if (CircDim == 2)
		{
			if (FindCP_2D(SP_ABS, AuxPos, EndPos, CP_ABS) < 0)
			{
				ret = ECAT_ERR_MC_INVALID_PARAM;
				goto out_return;
			}
		}
		else
		{
			if (FindCP_3D(SP_ABS, AuxPos, EndPos, CP_ABS, NV) < 0)
			{
				ret = ECAT_ERR_MC_INVALID_PARAM;
				goto out_return;
			}
		}

		for (i = 0; i < CircDim; i++)
		{
			SP_Rel_CP[i] = (SP_ABS[i] - CP_ABS[i]);
			EP_Rel_CP[i] = EndPos[i] - CP_ABS[i];
			R = R + (SP_Rel_CP[i] * SP_Rel_CP[i]);

			if (AbsMove)
				AuxPos[i] = CP_ABS[i];
			else
				AuxPos[i] = -SP_Rel_CP[i];
		}

		R = sqrt(R);
		if (R <= MC_DOUBLE_EQUAL_ZERO)
		{
			ret = ECAT_ERR_MC_INVALID_RADIUS;
			goto out_return;
		}
	}
	else if (CircMode == MC_CIRC_MODE_CP_EP)
	{
		for (i = 0; i < CircDim; i++)
		{
			if (AbsMove)
			{
				SP_Rel_CP[i] = (SP_ABS[i] - AuxPos[i]);
				CP_ABS[i] = AuxPos[i];
				EP_ABS[i] = EndPos[i];
			}
			else
			{
				SP_Rel_CP[i] = -AuxPos[i];
				CP_ABS[i] = (SP_ABS[i] + AuxPos[i]);
				EP_ABS[i] = SP_ABS[i] + EndPos[i];
			}
			R = R + (SP_Rel_CP[i] * SP_Rel_CP[i]);

			EP_Rel_CP[i] = EP_ABS[i] - CP_ABS[i];
			EP_Ridus = EP_Ridus + (EP_Rel_CP[i] * EP_Rel_CP[i]);


		}

		if (CircDim == 3)
		{
			if (FindNV(SP_ABS, CP_ABS, EP_ABS, NV) < 0)
			{
				ret = ECAT_ERR_MC_INVALID_END_POS;
				goto out_return;
			}
		}

		R = sqrt(R);
		if (R <= MC_DOUBLE_EQUAL_ZERO)
		{
			ret = ECAT_ERR_MC_INVALID_RADIUS;
			goto out_return;
		}

		EP_Ridus = sqrt(EP_Ridus);
		if (fabs(R - EP_Ridus) >= 0.01)
		{
			ret = ECAT_ERR_MC_INVALID_END_POS;
			goto out_return;
		}
	}


	//Define Normal vector
	if (CircDim == 2)
	{
		NV[0] = 0.0;
		NV[1] = 0.0;
		NV[2] = 1.0;
	}
	if ((CircMode == MC_CIRC_MODE_CP_EP) || (CircMode == MC_CIRC_MODE_BP_EP))
	{

		for (i = 0; i < CircDim; i++)
			x_UnitVector[i] = SP_Rel_CP[i] / R;

		Vector_Cross(3, y_UnitVector, NV, x_UnitVector);
		x_PjtValue = Vector_Dot(3, x_UnitVector, EP_Rel_CP);
		y_PjtValue = Vector_Dot(3, y_UnitVector, EP_Rel_CP);
		Angle = atan2(y_PjtValue, x_PjtValue); // 0~180->0~pi,180~360->-pi~0

		if (Angle >= 0)
		{
			if (!Dir)
				Angle = -M_2PI + Angle;
		}
		else
		{
			if (Dir)
				Angle = M_2PI + Angle;
		}
		Angle = Angle * 180 / M_PI;
	}

	if (Angle < 0)
		Pitch = Pitch * -1;

	Ops.GroupNo = GroupNo;
	Ops.Vel = Vel;
	Ops.Angle = Angle;
	Ops.CircDim = CircDim;
	Ops.AbsMove = AbsMove;
	Ops.MoveHelixType = MoveHelixType;
	Ops.Pitch = Pitch;
	Ops.EndRadius = EndRadius;
	memcpy(Ops.DoubleData, AuxPos, sizeof(double) * 3);
	memcpy(Ops.NormalVector, NV, sizeof(Ops.NormalVector));

	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_MOVE_CIRC_HELIX,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);

out_return:
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveCircularAbs(uint16_t DeviceNo, uint16_t GroupNo, double Angle, double AuxPos[], double Vel)
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 2, true, MC_GRP_NOT_MOVE_HELIX,
		Vel, Angle, 0, 0, AuxPos, NULL, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveCircularAbs_CP_Angle(uint16_t DeviceNo, uint16_t GroupNo, double Vel, double Angle, double AuxPos[])
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 2, true, MC_GRP_NOT_MOVE_HELIX,
		Vel, Angle, 0, 0, AuxPos, NULL, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveCircularAbs_CP_EP(uint16_t DeviceNo, uint16_t GroupNo, double Vel, uint8_t Dir, double AuxPos[], double EndPos[])
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_EP, 2, true, MC_GRP_NOT_MOVE_HELIX,
		Vel, 0, Dir, 0, AuxPos, EndPos, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveCircularAbs_BP_EP(uint16_t DeviceNo, uint16_t GroupNo, double Vel, uint8_t Dir, double AuxPos[], double EndPos[])
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_BP_EP, 2, true, MC_GRP_NOT_MOVE_HELIX,
		Vel, 0, Dir, 0, AuxPos, EndPos, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMove3DCircularAbs_CP_Angle(uint16_t DeviceNo, uint16_t GroupNo, double Vel, double Angle, double AuxPos[], double NV[])
{
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 3, true, MC_GRP_NOT_MOVE_HELIX,
		Vel, Angle, 0, 0, AuxPos, NULL, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMove3DCircularAbs_CP_EP(uint16_t DeviceNo, uint16_t GroupNo, double Vel, uint8_t Dir, double AuxPos[], double EndPos[])
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_EP, 3, true, MC_GRP_NOT_MOVE_HELIX,
		Vel, 0, Dir, 0, AuxPos, EndPos, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMove3DCircularAbs_BP_EP(uint16_t DeviceNo, uint16_t GroupNo, double Vel, uint8_t Dir, double AuxPos[], double EndPos[])
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_BP_EP, 3, true, MC_GRP_NOT_MOVE_HELIX,
		Vel, 0, Dir, 0, AuxPos, EndPos, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveCircularRel(uint16_t DeviceNo, uint16_t GroupNo, double Angle, double AuxPos[], double Vel)
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 2, false, MC_GRP_NOT_MOVE_HELIX,
		Vel, Angle, 0, 0, AuxPos, NULL, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveCircularRel_CP_Angle(uint16_t DeviceNo, uint16_t GroupNo, double Vel, double Angle, double AuxPos[])
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 2, false, MC_GRP_NOT_MOVE_HELIX,
		Vel, Angle, 0, 0, AuxPos, NULL, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveCircularRel_CP_EP(uint16_t DeviceNo, uint16_t GroupNo, double Vel, uint8_t Dir, double AuxPos[], double EndPos[])
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_EP, 2, false, MC_GRP_NOT_MOVE_HELIX,
		Vel, 0, Dir, 0, AuxPos, EndPos, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveCircularRel_BP_EP(uint16_t DeviceNo, uint16_t GroupNo, double Vel, uint8_t Dir, double AuxPos[], double EndPos[])
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_BP_EP, 2, false, MC_GRP_NOT_MOVE_HELIX,
		Vel, 0, Dir, 0, AuxPos, EndPos, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMove3DCircularRel_CP_Angle(uint16_t DeviceNo, uint16_t GroupNo, double Vel, double Angle, double AuxPos[], double NV[])
{
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 3, false, MC_GRP_NOT_MOVE_HELIX,
		Vel, Angle, 0, 0, AuxPos, NULL, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMove3DCircularRel_CP_EP(uint16_t DeviceNo, uint16_t GroupNo, double Vel, uint8_t Dir, double AuxPos[], double EndPos[])
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_EP, 3, false, MC_GRP_NOT_MOVE_HELIX,
		Vel, 0, Dir, 0, AuxPos, EndPos, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMove3DCircularRel_BP_EP(uint16_t DeviceNo, uint16_t GroupNo, double Vel, uint8_t Dir, double AuxPos[], double EndPos[])
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_BP_EP, 3, false, MC_GRP_NOT_MOVE_HELIX,
		Vel, 0, Dir, 0, AuxPos, EndPos, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveHelicalAbs(uint16_t DeviceNo, uint16_t GroupNo, double Angle, double AuxPos[], double Pitch, double Vel)
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 2, true, MC_GRP_MOVE_NORMAL_HELIX,
		Vel, Angle, 0, Pitch, AuxPos, NULL, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMove3DHelicalAbs_CP_Angle(uint16_t DeviceNo, uint16_t GroupNo, double Vel, double Angle, double Pitch, double AuxPos[], double NV[])
{
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 3, true, MC_GRP_MOVE_NORMAL_HELIX,
		Vel, Angle, 0, Pitch, AuxPos, NULL, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveHelicalRel(uint16_t DeviceNo, uint16_t GroupNo, double Angle, double AuxPos[], double Pitch, double Vel)
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 2, false, MC_GRP_MOVE_NORMAL_HELIX,
		Vel, Angle, 0, Pitch, AuxPos, NULL, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMove3DHelicalRel_CP_Angle(uint16_t DeviceNo, uint16_t GroupNo, double Vel, double Angle, double Pitch, double AuxPos[], double NV[])
{
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 3, false, MC_GRP_MOVE_NORMAL_HELIX,
		Vel, Angle, 0, Pitch, AuxPos, NULL, NV, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveConicalHelixAbs(uint16_t DeviceNo, uint16_t GroupNo, double Angle, double AuxPos[], double Pitch, double Vel, double EndRadius)
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 2, true, MC_GRP_MOVE_CONICAL_HELIX,
		Vel, Angle, 0, Pitch, AuxPos, NULL, NV, EndRadius);
}
/*****************************************************************************/
int32_t ECAT_McGroupMove3DConicalHelixAbs_CP_Angle(uint16_t DeviceNo, uint16_t GroupNo, double Vel, double Angle, double Pitch, double AuxPos[], double NV[], double EndRadius)
{
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 3, true, MC_GRP_MOVE_CONICAL_HELIX,
		Vel, Angle, 0, Pitch, AuxPos, NULL, NV, EndRadius);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveConicalHelixRel(uint16_t DeviceNo, uint16_t GroupNo, double Angle, double AuxPos[], double Pitch, double Vel, double EndRadius)
{
	double NV[3];
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 2, false, MC_GRP_MOVE_CONICAL_HELIX,
		Vel, Angle, 0, Pitch, AuxPos, NULL, NV, EndRadius);
}
/*****************************************************************************/
int32_t ECAT_McGroupMove3DConicalHelixRel_CP_Angle(uint16_t DeviceNo, uint16_t GroupNo, double Vel, double Angle, double Pitch, double AuxPos[], double NV[], double EndRadius)
{
	return SendMcGroupMoveCircHelix(DeviceNo, GroupNo, MC_CIRC_MODE_CP_ANGLE, 3, false, MC_GRP_MOVE_CONICAL_HELIX,
		Vel, Angle, 0, Pitch, AuxPos, NULL, NV, EndRadius);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveProfile(uint16_t DeviceNo, uint16_t GroupNo, uint16_t ProfileNo[], uint16_t TotalStep)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = TotalStep;
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	memcpy(Ops.ProfileNo, ProfileNo, sizeof(uint16_t)*EcatDevice[DeviceNo].GrpAxisCnt[GroupNo]);
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_MOVE_PROFILE,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveProfileCSV(uint16_t DeviceNo, uint16_t GroupNo, uint16_t ProfileNo)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.ProfileNo[0] = ProfileNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_MOVE_PROFILE_CSV,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveDwell(uint16_t DeviceNo, uint16_t GroupNo, uint32_t Cnt)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U32Data = Cnt;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_MOVE_DWELL,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveDO(uint16_t DeviceNo, uint16_t GroupNo, uint16_t SlaveNo, uint16_t BitNo, uint32_t Value)
{
	ecat_mc_group_move_do_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.SlaveNo = SlaveNo;
	Ops.BitNo = BitNo;
	Ops.Value = Value;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_MOVE_DO,
							&Ops, sizeof(ecat_mc_group_move_do_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveBlendingSync(uint16_t DeviceNo, uint16_t GroupNo)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_MOVE_BLENDING_SYNC,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_SetDeviceEncProperty(uint16_t DeviceNo, uint16_t EncNo, uint8_t Mode, uint8_t InvertCnt, uint8_t LPF)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	if (EncNo >= DEV_ENCODER_MAX)
		return ECAT_ERR_MC_INVALID_AXIS_NO;

	int32_t ret;
	encoder_config_t config;
	uint32_t offset;
	if (EncNo == 0)
		offset = ENC0_MODE_ADDR_OFFSET;
	else
		offset = ENC1_MODE_ADDR_OFFSET;
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, offset, &config.Cfg, 1) != 0))
		goto out_return;
	config.Mode0 = InvertCnt;
	config.EncoderMode = Mode;
	config.LPF = LPF;
	ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, offset, &config.Cfg, 1);
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceEncProperty(uint16_t DeviceNo, uint16_t EncNo, uint8_t *Mode, uint8_t *InvertCnt, uint8_t *LPF)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	if (EncNo >= DEV_ENCODER_MAX)
		return ECAT_ERR_MC_INVALID_AXIS_NO;

	int32_t ret;
	encoder_config_t config;
	uint32_t offset;
	if (EncNo == 0)
		offset = ENC0_MODE_ADDR_OFFSET;
	else
		offset = ENC1_MODE_ADDR_OFFSET;
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, offset, &config.Cfg, 1) != 0))
		goto out_return;
	*InvertCnt = config.Mode0;
	*Mode = config.EncoderMode;
	*LPF = config.LPF;
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceEnCount(uint16_t DeviceNo, uint16_t EncNo, int32_t *Cnt)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	if (EncNo >= DEV_ENCODER_MAX)
		return ECAT_ERR_MC_INVALID_AXIS_NO;

	int32_t ret;
	uint32_t offset;
	int32_t enc_cnt = 0;
	if (EncNo == 0)
		offset = ENC0_CNT_ADDR_OFFSET;
	else
		offset = ENC1_CNT_ADDR_OFFSET;
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, offset, &enc_cnt, 1) != 0))
		goto out_return;
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, offset, &enc_cnt, 1) != 0))
		goto out_return;
	*Cnt = enc_cnt;
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_ResetDeviceEncCount(uint16_t DeviceNo, uint16_t EncNo)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	if (EncNo >= DEV_ENCODER_MAX)
		return ECAT_ERR_MC_INVALID_AXIS_NO;

	int32_t ret;
	encoder_config_t config;
	uint32_t offset;
	if (EncNo == 0)
		offset = ENC0_MODE_ADDR_OFFSET;
	else
		offset = ENC1_MODE_ADDR_OFFSET;
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, offset, &config.Cfg, 1) != 0))
		goto out_return;
	ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, offset, &config.Cfg, 1);
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetDeviceCmpTrigProperty(uint16_t DeviceNo, uint16_t EncNo, uint32_t PulseWidth, uint8_t Source)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	if (EncNo >= DEV_ENCODER_MAX)
		return ECAT_ERR_MC_INVALID_AXIS_NO;

	int32_t ret;
	uint32_t config;
	uint32_t offset;
	if (EncNo == 0)
		offset = ENC0_CMP_TRIG_CONFIG_ADDR_OFFSET;
	else
		offset = ENC1_CMP_TRIG_CONFIG_ADDR_OFFSET;
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, offset, &config, 1) != 0))
		goto out_return;
	PulseWidth |= 0x80000000;
	config &= 0x80000000;
	config |= PulseWidth;
	if ((ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, offset, &config, 1) != 0))
		goto out_return;

	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, CMP_TRIG_COMM_CONFIG_ADDR_OFFSET, &config, 1) != 0))
		goto out_return;
	if (Source)
		config |= 0x10;
	else
		config &= ~0x10;
	ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, CMP_TRIG_COMM_CONFIG_ADDR_OFFSET, &config, 1);
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceCmpTrigProperty(uint16_t DeviceNo, uint16_t EncNo, uint32_t *PulseWidth, uint8_t *Source)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	if (EncNo >= DEV_ENCODER_MAX)
		return ECAT_ERR_MC_INVALID_AXIS_NO;

	int32_t ret;
	uint32_t config;
	uint32_t offset;
	if (EncNo == 0)
		offset = ENC0_CMP_TRIG_CONFIG_ADDR_OFFSET;
	else
		offset = ENC1_CMP_TRIG_CONFIG_ADDR_OFFSET;
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, offset, &config, 1) != 0))
		goto out_return;
	*PulseWidth = (config & 0x7FFFFFFF);

	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, CMP_TRIG_COMM_CONFIG_ADDR_OFFSET, &config, 1) != 0))
		goto out_return;
	config &= 0x10;
	if (config)
		*Source = 1;
	else
		*Source = 0;
out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t __SetDeviceCmpTrigData(uint16_t DeviceNo, uint16_t EncNo, int32_t CmpData, uint32_t Interval, uint32_t Times, uint8_t Dir)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	if (EncNo >= DEV_ENCODER_MAX)
		return ECAT_ERR_MC_INVALID_AXIS_NO;

	int32_t ret;
	uint32_t config;
	uint32_t config_offset, data_offset, times_offset, pitch_offset;
	if (EncNo == 0) {
		config_offset = ENC0_CMP_TRIG_CONFIG_ADDR_OFFSET;
		data_offset = ENC0_CMP_TRIG_DATA_ADDR_OFFSET;
		times_offset = ENC0_CMP_TRIG_TIMES_ADDR_OFFSET;
		pitch_offset = ENC0_CMP_TRIG_PITCH_ADDR_OFFSET;
	} else {
		config_offset = ENC1_CMP_TRIG_CONFIG_ADDR_OFFSET;
		data_offset = ENC1_CMP_TRIG_DATA_ADDR_OFFSET;
		times_offset = ENC1_CMP_TRIG_TIMES_ADDR_OFFSET;
		pitch_offset = ENC1_CMP_TRIG_PITCH_ADDR_OFFSET;
	}

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, CMP_TRIG_COMM_CONFIG_ADDR_OFFSET, &config, 1) != 0))
		goto out_return;
	if (Dir)
		config |= (0x01 << EncNo);
	else
		config &= ~(0x01 << EncNo);
	if ((ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, CMP_TRIG_COMM_CONFIG_ADDR_OFFSET, &config, 1) != 0))
		goto out_return;

	if ((ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, data_offset, &CmpData, 1) != 0))
		goto out_return;

	if (Times == 1) {
		config = 0;
		if ((ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, pitch_offset, &config, 1) != 0))
			goto out_return;
	} else {
		if ((ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, pitch_offset, &Interval, 1) != 0))
			goto out_return;
	}

	if (Times)
			config = 0x80000000 | Times;
		else
			config = 0;
	if ((ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, times_offset, &config, 1) != 0))
		goto out_return;

	if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, config_offset, &config, 1) != 0))
		goto out_return;
	config &= 0x7FFFFFFF;
	if ((ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, config_offset, &config, 1) != 0))
		goto out_return;
	config |= 0x80000000;
	if ((ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, config_offset, &config, 1) != 0))
		goto out_return;

out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetDeviceCmpTrigData(uint16_t DeviceNo, uint16_t EncNo, int32_t CmpData)
{
	return __SetDeviceCmpTrigData(DeviceNo, EncNo, CmpData, 0, 1, 0);
}
/*****************************************************************************/
int32_t ECAT_SetDeviceContCmpTrigData(uint16_t DeviceNo, uint16_t EncNo, int32_t Start, uint32_t Interval, uint32_t Times, uint8_t Dir)
{
	return __SetDeviceCmpTrigData(DeviceNo, EncNo, Start, Interval, Times, Dir);
}
/*****************************************************************************/
int32_t ECAT_SetDeviceEmg(uint16_t DeviceNo, uint8_t Source, uint8_t Enable, uint8_t Logic, uint16_t SlaveNo, uint16_t BitNo, uint8_t ServoOff)
{
	ecat_device_emg_t Ops;
	Ops.Source = Source;
	Ops.Enable = Enable;
	Ops.Logic = Logic;
	Ops.SlaveNo = SlaveNo;
	Ops.BitNo = BitNo;
	Ops.ServoOff = ServoOff;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_DEVICE_EMG,
							&Ops, sizeof(ecat_device_emg_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_GetDeviceEmg(uint16_t DeviceNo, uint8_t *Source, uint8_t *Enable, uint8_t *Logic, uint16_t *SlaveNo, uint16_t *BitNo, uint8_t *ServoOff)
{
	int32_t ret;
	ecat_device_emg_t RetData;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_DEVICE_EMG,
							NULL, 0, &RetData, sizeof(ecat_device_emg_t));
	if (ret == 0) {
		*Source = RetData.Source;
		*Enable = RetData.Enable;
		*Logic = RetData.Logic;
		*SlaveNo = RetData.SlaveNo;
		*BitNo = RetData.BitNo;
		*ServoOff = RetData.ServoOff;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceEmgStatus(uint16_t DeviceNo, uint8_t *Status)
{
	int32_t ret;
	ecat_device_emg_t RetData;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_DEVICE_EMG_STATUS,
							NULL, 0, &RetData, sizeof(ecat_device_emg_t));
	if (ret == 0) {
		*Status = RetData.Status;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetDeviceEmgSoftSig(uint16_t DeviceNo)
{
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_DEVICE_EMG_SIG,
							NULL, 0, NULL, 0);
}
/*****************************************************************************/
void* ThreadMpgFunc(void* arg)
{
	printf("ThreadMpgFunc Start\n");
	ecat_device_ipc_t *EcatDevice = (ecat_device_ipc_t *)arg;
	int32_t NewCnt = 0, OldCnt = 0, DiffCnt = 0;
	uint16_t OldAxisNo, AxisNo;
	uint32_t Scaling = 0, OldScaling = 0;
	uint32_t DI = 0, config;
	ecat_mc_axis_ops_t Ops;
	uint32_t OpsSize = sizeof(ecat_mc_axis_ops_t);
	Ops.AxisNo = 0;
	EcatDevice->MpgThreadRunning = 1;


	while (!EcatDevice->MpgThreadShouldStop) {
		SEMEPHORE_P_OPERATION(EcatDevice->ThreadSem)
		if (read_bar0_reg(EcatDevice->Fd, GPIO_ADDR_OFFSET, &DI, 1) != 0)
			goto sleep;

		DI = ~(DI >> 16);
		DI &= 0x1FFF;
		if (DI & ((1 << EcatDevice->MpgAxisCnt) - 1)) {
			AxisNo = 0;
			Ops.AxisNo = 0;
			Scaling = 1;
			do {
				if (DI & (1 << AxisNo)) {
					Ops.AxisNo = EcatDevice->MpgAxisNo[AxisNo];
					break;
				}
				else
					AxisNo++;

				if (AxisNo >= EcatDevice->MpgAxisCnt)
					break;

			} while (1);

			if (DI & 0x80)
				Scaling = 1;
			else if (DI & 0x100)
				Scaling = 10;
			else if (DI & 0x200)
				Scaling = 100;

			if ((Scaling != OldScaling) || (Ops.AxisNo != OldAxisNo)) {
				if (read_bar0_reg(EcatDevice->Fd, ENC0_MODE_ADDR_OFFSET, &config, 1) != 0)
					goto sleep;
				if (write_bar0_reg(EcatDevice->Fd, ENC0_MODE_ADDR_OFFSET, &config, 1) != 0)
					goto sleep;
				OldScaling = Scaling;
				OldAxisNo = Ops.AxisNo;
				NewCnt = 0;
				OldCnt = 0;
			}

			NewCnt = 0;
			if (write_bar0_reg(EcatDevice->Fd, ENC0_CNT_ADDR_OFFSET, &NewCnt, 1) != 0)
				goto sleep;
			if (read_bar0_reg(EcatDevice->Fd, ENC0_CNT_ADDR_OFFSET, &NewCnt, 1) != 0)
				goto sleep;
			DiffCnt = NewCnt - OldCnt;
			if (DiffCnt != 0) 	{
				OldCnt = NewCnt;
				Ops.I32Data = DiffCnt * Scaling;
				SetCommandData(EcatDevice, ECAT_IPC_CMD_MC_AXIS_MOVE_MPG, (uint32_t*)&Ops, OpsSize);
			}
		}
		else
			OldAxisNo = 0xFFFF;

sleep:
		SEMEPHORE_V_OPERATION(EcatDevice->ThreadSem)
		usleep(1000);
	}
	SEMEPHORE_P_OPERATION(EcatDevice->ThreadSem)
	EcatDevice->MpgThreadShouldStop = 0;
	EcatDevice->MpgThreadRunning = 0;
	SEMEPHORE_V_OPERATION(EcatDevice->ThreadSem)
	printf("ThreadMpgFunc Stop\n");
	pthread_exit(0);
}
/*****************************************************************************/
int32_t ECAT_SetDeviceMPG(uint16_t DeviceNo, uint8_t Enable, uint16_t *AxisNo, uint16_t AxisCount)
{
	uint32_t i;
	int32_t ret = 0;
	encoder_config_t config;
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)

	if (Enable) {
		if (!EcatDevice[DeviceNo].MpgThreadRunning) {
			if ((AxisCount) && (AxisCount >= MC_MPG_AXIS_NO_MAX))
				return ECAT_ERR_MC_INVALID_PARAM;

			SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
			if ((AxisCount) && (AxisNo != NULL)) {
				memcpy(EcatDevice[DeviceNo].MpgAxisNo, AxisNo, sizeof(uint16_t) * AxisCount);
				EcatDevice[DeviceNo].MpgAxisCnt = AxisCount;
			} else {
				for (i = 0; i < MC_MPG_AXIS_NO_MAX; i++)
					EcatDevice[DeviceNo].MpgAxisNo[i] = i;
				EcatDevice[DeviceNo].MpgAxisCnt = 4;
			}

			if ((ret = read_bar0_reg(EcatDevice[DeviceNo].Fd, ENC0_MODE_ADDR_OFFSET, &config.Cfg, 1) != 0))
				goto out_return;
			config.Mode0 = 0;
			config.EncoderMode = DEV_ENC_MODE_AB_PHASE;
			config.LPF = 0;
			if ((ret = write_bar0_reg(EcatDevice[DeviceNo].Fd, ENC0_MODE_ADDR_OFFSET, &config.Cfg, 1) != 0))
				goto out_return;

			EcatDevice[DeviceNo].MpgThreadShouldStop = 0;
			if ((pthread_create(&EcatDevice[DeviceNo].MpgThreadID, NULL,
					ThreadMpgFunc, (void*)&EcatDevice[DeviceNo]) != 0)) {
					ret = ECAT_ERR_IPC_CREATE_HANDLE;
			}
out_return:
			SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
			return ret;
		}
	} else {
		if ((!EcatDevice[DeviceNo].MpgThreadShouldStop) && (EcatDevice[DeviceNo].MpgThreadRunning)) {
			SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
			EcatDevice[DeviceNo].MpgThreadShouldStop = 1;
			SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
			pthread_join(EcatDevice[DeviceNo].MpgThreadID, NULL);

		}
	}
	return 0;
}
/*****************************************************************************/
int32_t ECAT_GetDeviceMPG(uint16_t DeviceNo, uint8_t *Enable, uint16_t *AxisNo, uint16_t *AxisCount)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if (EcatDevice[DeviceNo].MpgThreadRunning)
		*Enable = 0;
	else
		*Enable = 1;

	if (AxisCount)
		*AxisCount = EcatDevice[DeviceNo].MpgAxisCnt;
	if (AxisNo)
		memcpy(AxisNo, EcatDevice[DeviceNo].MpgAxisNo, sizeof(uint16_t) * EcatDevice[DeviceNo].MpgAxisCnt);
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return 0;

}
/*****************************************************************************/
int32_t ECAT_GetSlaveRegister(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t Addr, uint16_t DataSize, uint32_t *Val, uint8_t *state)
{
	int32_t ret;
	ecat_slave_reg_ops_t Ops;
	ecat_slave_reg_return_t RetData;
	Ops.No = SlaveNo;
	Ops.Addr = Addr;
	Ops.DataSize = DataSize;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_REG_VALUE,
							&Ops, sizeof(ecat_slave_reg_ops_t), &RetData, sizeof(ecat_slave_reg_return_t));
	if (ret == 0)
		*Val = RetData.Val;
	else
		*state = RetData.State;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetSlaveRegister(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t Addr, uint16_t DataSize, uint32_t Val, uint8_t *state)
{
	int32_t ret;
	ecat_slave_reg_ops_t Ops;
	ecat_slave_reg_return_t RetData;
	Ops.No = SlaveNo;
	Ops.Addr = Addr;
	Ops.DataSize = DataSize;
	Ops.Val = Val;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_REG_VALUE,
							&Ops, sizeof(ecat_slave_reg_ops_t), &RetData, sizeof(ecat_slave_reg_return_t));
	if (ret != 0) {
		*state = RetData.State;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetMotionRecordState(uint16_t DeviceNo, uint16_t *state, uint32_t *count)
{
	int32_t ret;
	ecat_mc_return_t RetData;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_RECORD_STATE,
							NULL, 0, &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0) {
		*state = RetData.U16Value;
		*count = RetData.U32Value;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetMotionRecord(uint16_t DeviceNo, uint16_t state)
{
	ecat_ipc_base_ops_t Ops;
	Ops.No = state;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_RECORD,
							&Ops, sizeof(ecat_ipc_base_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_DownloadMotionLog(uint16_t DeviceNo, uint32_t *Offset, char *Data, uint32_t *DataSize, uint8_t *LastFlag)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret;
	ecat_read_file_ops_t Ops;
	ecat_read_file_return_t RetData;
	uint32_t OpsSize = sizeof(ecat_read_file_ops_t);
	uint32_t RetSize = sizeof(ecat_read_file_return_t);
	uint32_t DataOffset = 0;

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if (*LastFlag == 255)
	{
		Ops.GetFileSize = 1;


		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_MC_DOWNLOAD_MOTION_LOG, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		if ((ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&RetData, RetSize)) < 0)
			goto out_return;
		else
		{
			*DataSize = RetData.Size;
			ret = 0;
			goto out_return;
		}

	}
	else
	{
		Ops.GetFileSize = 0;
		Ops.offset = *Offset;

		if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_MC_DOWNLOAD_MOTION_LOG, (uint32_t*)&Ops, OpsSize)) < 0)
			goto out_return;

		if ((ret = GetReturnData(&EcatDevice[DeviceNo], (uint32_t*)&RetData, RetSize)) < 0)
			goto out_return;
		else
		{
			memcpy(Data, RetData.Data, RetData.Size);
			*DataSize = RetData.Size;
			*Offset = RetData.Offset;
			*LastFlag = RetData.LastFlag;
			ret = 0;
			goto out_return;
		}
	}

out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return ret;
}
/*****************************************************************************/
int32_t timer_setup(uint16_t DeviceNo, uint32_t Interval)
{
	int32_t ret;
	ecat_u32_t Ops;
	Ops.u32data = Interval;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_TIMER,
							&Ops, sizeof(ecat_u32_t), NULL, 0);
	if ((ret == 0) && (Interval)) {
		ret = irq_setup(DeviceNo, 0x01);
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetTimer(uint16_t DeviceNo, uint32_t Interval)
{
	return timer_setup(DeviceNo, Interval);
}
/*****************************************************************************/
int32_t ECAT_SetTimerStop(uint16_t DeviceNo)
{
	return timer_setup(DeviceNo, 0);
}
/*****************************************************************************/
int32_t ECAT_WaitforTimer(uint16_t DeviceNo)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int ret = 0;

	pthread_mutex_lock(&EcatDevice[DeviceNo].cond_timer_mutex);

	if (pthread_cond_wait(&EcatDevice[DeviceNo].cond_timer, &EcatDevice[DeviceNo].cond_timer_mutex) != 0)
		ret = ECAT_ERR_DRV_IOCTL;


	pthread_mutex_unlock(&EcatDevice[DeviceNo].cond_timer_mutex);
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetStewartPlatform_M1(uint16_t DeviceNo, double radiusB, double angleB, double radiusP, double angleP, double RodLength, double Max_RodLength)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int32_t ret;
	ecat_stewart_t Ops;
	ecat_slave_reg_return_t RetData;
	uint32_t OpsSize = sizeof(ecat_stewart_t);
	uint32_t RetSize = sizeof(ecat_slave_reg_return_t);

	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	stewartPlatform_t *Sp = &EcatDevice[DeviceNo].StewartPlatform;
	int i;
	Sp->rb = radiusB;
	Sp->rp = radiusP;
	Sp->ab = angleB / 180 * M_PI;
	Sp->ap = angleP / 180 * M_PI;
	Sp->ab_degree = angleB;
	Sp->ap_degree = angleP;

	for (i = 0; i<6; i++)
	{
		Sp->L0[i] = RodLength;
	}
	Sp->Z0 = sqrt(Sp->L0[0] * Sp->L0[0] - (Sp->rb - Sp->rp)*(Sp->rb - Sp->rp));
	/******************************/

	i = 1;
	Sp->angP[i] = i*M_PI / 3 - 0.5*Sp->ap;
	Sp->Px[i] = Sp->rp * cos(Sp->angP[i]);
	Sp->Py[i] = Sp->rp * sin(Sp->angP[i]);
	Sp->Pz[i] = 0;
	i = 2;
	Sp->angP[i] = Sp->angP[i - 1] + Sp->ap;
	Sp->Px[i] = Sp->rp * cos(Sp->angP[i]);
	Sp->Py[i] = Sp->rp * sin(Sp->angP[i]);
	Sp->Pz[i] = 0;
	i = 3;
	Sp->angP[i] = i*M_PI / 3 - 0.5*Sp->ap;
	Sp->Px[i] = Sp->rp * cos(Sp->angP[i]);
	Sp->Py[i] = Sp->rp * sin(Sp->angP[i]);
	Sp->Pz[i] = 0;
	i = 4;
	Sp->angP[i] = Sp->angP[i - 1] + Sp->ap;
	Sp->Px[i] = Sp->rp * cos(Sp->angP[i]);
	Sp->Py[i] = Sp->rp * sin(Sp->angP[i]);
	Sp->Pz[i] = 0;
	i = 5;
	Sp->angP[i] = i*M_PI / 3 - 0.5*Sp->ap;
	Sp->Px[i] = Sp->rp * cos(Sp->angP[i]);
	Sp->Py[i] = Sp->rp * sin(Sp->angP[i]);
	Sp->Pz[i] = 0;
	i = 6;
	Sp->angP[i] = Sp->angP[i - 1] + Sp->ap;
	Sp->Px[i] = Sp->rp * cos(Sp->angP[i]);
	Sp->Py[i] = Sp->rp * sin(Sp->angP[i]);
	Sp->Pz[i] = 0;

	/******************************/
	i = 1;
	Sp->angB[i] = i*M_PI / 3 - 0.5*Sp->ab;
	Sp->Bx[i] = Sp->rb * cos(Sp->angB[i]);
	Sp->By[i] = Sp->rb * sin(Sp->angB[i]);
	Sp->Bz[i] = 0;
	i = 2;
	Sp->angB[i] = Sp->angB[i - 1] + Sp->ab;
	Sp->Bx[i] = Sp->rb * cos(Sp->angB[i]);
	Sp->By[i] = Sp->rb * sin(Sp->angB[i]);
	Sp->Bz[i] = 0;
	i = 3;
	Sp->angB[i] = i*M_PI / 3 - 0.5*Sp->ab;
	Sp->Bx[i] = Sp->rb * cos(Sp->angB[i]);
	Sp->By[i] = Sp->rb * sin(Sp->angB[i]);
	Sp->Bz[i] = 0;
	i = 4;
	Sp->angB[i] = Sp->angB[i - 1] + Sp->ab;
	Sp->Bx[i] = Sp->rb * cos(Sp->angB[i]);
	Sp->By[i] = Sp->rb * sin(Sp->angB[i]);
	Sp->Bz[i] = 0;
	i = 5;
	Sp->angB[i] = i*M_PI / 3 - 0.5*Sp->ab;
	Sp->Bx[i] = Sp->rb * cos(Sp->angB[i]);
	Sp->By[i] = Sp->rb * sin(Sp->angB[i]);
	Sp->Bz[i] = 0;
	i = 6;
	Sp->angB[i] = Sp->angB[i - 1] + Sp->ab;
	Sp->Bx[i] = Sp->rb * cos(Sp->angB[i]);
	Sp->By[i] = Sp->rb * sin(Sp->angB[i]);
	Sp->Bz[i] = 0;

	for (i = 0; i<6; i++)
	{
		Sp->L0[i] = RodLength;
		Sp->Lmin[i] = RodLength;
		Sp->Lmax[i] = Max_RodLength;
	}
	for (i = 0; i<6; i++)
	{
		Ops.Bx[i] = Sp->Bx[i + 1];
		Ops.By[i] = Sp->By[i + 1];
		Ops.Px[i] = Sp->Px[i + 1];
		Ops.Py[i] = Sp->Py[i + 1];
	}
	Ops.Z0 = Sp->Z0;
	memcpy(Ops.Max_RodLength, Sp->Lmax, sizeof(Sp->Lmax));
	memcpy(Ops.Min_RodLength, Sp->Lmin, sizeof(Sp->Lmin));
	if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_SET_STEWART_PLATFORM, (uint32_t*)&Ops, OpsSize)) < 0)
		goto out_return;

	ret = GetReturnData(&EcatDevice[DeviceNo], NULL, 0);

out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if (ret == 0)
		Sp->setup = 1;

	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisActualPosVel(uint16_t DeviceNo, uint16_t AxisNo, float *Pos, float *Vel)
{
	int32_t ret;
	int i;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_float_multi_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_ACT_POS_VEL,
							&Ops, sizeof(ecat_mc_axis_ops_t), &RetData, sizeof(ecat_mc_return_float_multi_t));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++) {
				*(Pos + i) = RetData.FloatValue_multi_Value1[i];
				*(Vel + i) = RetData.FloatValue_multi_Value2[i];
			}
		} else {
			*Pos = RetData.FloatValue_multi_Value1[0];
			*Vel = RetData.FloatValue_multi_Value2[0];
		}
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidSetSetPointValue(uint16_t DeviceNo, uint32_t PidNo, double SetPointValue)
{
	Pid_ecat_double_t Ops;
	Ops.PidNo = PidNo;
	Ops.doubledata = SetPointValue;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_PID_SET_POINT_VALUE,
							&Ops, sizeof(Pid_ecat_double_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_PidGetControlOutputValue(uint16_t DeviceNo, uint32_t PidNo, double *Output)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	ecat_pid_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_PARAMETER,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(ecat_pid_t));
	if (ret == 0) {
		*Output = RetData.Feedback;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidSetParameter(uint16_t DeviceNo, uint32_t PidNo, double kp, double ki, double kd)
{
	ecat_pid_t Ops;
	Ops.PidNo = PidNo;
	Ops.kp = kp;
	Ops.ki = ki;
	Ops.kd = kd;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_PID_PARAMETER,
							&Ops, sizeof(ecat_pid_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_PidSetStatus(uint16_t DeviceNo, uint32_t PidNo, uint8_t status)
{
	ecat_pid_t Ops;
	Ops.PidNo = PidNo;
	Ops.status = status;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_PID_STATUS,
							&Ops, sizeof(ecat_pid_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_PidSetSimulateMode(uint16_t DeviceNo, uint32_t PidNo, uint8_t status)
{
	ecat_pid_t Ops;
	Ops.PidNo = PidNo;
	Ops.status = status;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_PID_SIMULATE,
							&Ops, sizeof(ecat_pid_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_PidSetSampleTime(uint16_t DeviceNo, uint32_t PidNo, uint32_t Interval)
{
	ecat_pid_controller_t Ops;
	Ops.PidNo = PidNo;
	Ops.Interval = Interval;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_PID_CONTROLLER,
							&Ops, sizeof(ecat_pid_controller_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_PidSetProcessVariableModule(uint16_t DeviceNo, uint32_t PidNo, uint16_t SlaveNo, uint16_t OffsetByte, uint16_t Bitlength, double ScaleGain, double ScaleOffset)
{
	ecat_pid_controller_t Ops;
	Ops.PidNo = PidNo;
	Ops.SlaveNo_In = SlaveNo;
	Ops.OffsetByte_In = OffsetByte;
	Ops.Bitlength_In = Bitlength;
	Ops.Scalegain_In = ScaleGain;
	Ops.Scaleoffset_In = ScaleOffset;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_PID_PROCESS_VARIABLE_MODULE,
							&Ops, sizeof(ecat_pid_controller_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_PidSetControlOutputModule(uint16_t DeviceNo, uint32_t PidNo, uint16_t SlaveNo, uint16_t OffsetByte, uint16_t Bitlength, double ScaleGain, double ScaleOffset, double Output_Max_Value, double Output_Min_Value)
{
	ecat_pid_controller_t Ops;
	Ops.PidNo = PidNo;
	Ops.SlaveNo_Out = SlaveNo;
	Ops.OffsetByte_Out = OffsetByte;
	Ops.Bitlength_Out = Bitlength;
	Ops.Scalegain_Out = ScaleGain;
	Ops.Scaleoffset_Out = ScaleOffset;
	Ops.Output_max = Output_Max_Value;
	Ops.Output_min = Output_Min_Value;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_PID_CONTROL_OUTPUT_MODULE,
							&Ops, sizeof(ecat_pid_controller_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_PidGetParameter(uint16_t DeviceNo, uint32_t PidNo, double *kp, double *ki, double *kd)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	ecat_pid_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_PARAMETER,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(ecat_pid_t));
	if (ret == 0) {
		*kp = RetData.kp;
		*ki = RetData.ki;
		*kd = RetData.kd;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidGetStatus(uint16_t DeviceNo, uint32_t PidNo, uint8_t *status)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	ecat_pid_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_STATUS,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(ecat_pid_t));
	if (ret == 0) {
		*status = RetData.status;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidGetSimulateMode(uint16_t DeviceNo, uint32_t PidNo, uint8_t *status)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	ecat_pid_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_SIMULATE,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(ecat_pid_t));
	if (ret == 0) {
		*status = RetData.status;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidGetSampleTime(uint16_t DeviceNo, uint32_t PidNo, uint32_t *Interval)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	ecat_pid_controller_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_CONTROLLER,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(ecat_pid_controller_t));
	if (ret == 0) {
		*Interval = RetData.Interval;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidGetProcessVariableModule(uint16_t DeviceNo, uint32_t PidNo, uint16_t *SlaveNo, uint16_t *OffsetByte, uint16_t *Bitlength, double *ScaleGain, double *ScaleOffset)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	ecat_pid_controller_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_PROCESS_VARIABLE_MODULE,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(ecat_pid_controller_t));
	if (ret == 0) {
		*SlaveNo = RetData.SlaveNo_In;
		*OffsetByte = RetData.OffsetByte_In;
		*Bitlength = RetData.Bitlength_In;
		*ScaleGain = RetData.Scalegain_In;
		*ScaleOffset = RetData.Scaleoffset_In;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidGetControlOutputModule(uint16_t DeviceNo, uint32_t PidNo, uint16_t *SlaveNo, uint16_t *OffsetByte, uint16_t *Bitlength, double *ScaleGain, double *ScaleOffset, double *Output_Max_Value, double *Output_Min_Value)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	ecat_pid_controller_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_CONTROL_OUTPUT_MODULE,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(ecat_pid_controller_t));
	if (ret == 0) {
		*SlaveNo = RetData.SlaveNo_Out;
		*OffsetByte = RetData.OffsetByte_Out;
		*Bitlength = RetData.Bitlength_Out;
		*ScaleGain = RetData.Scalegain_Out;
		*ScaleOffset = RetData.Scaleoffset_Out;
		*Output_Max_Value = RetData.Output_max;
		*Output_Min_Value = RetData.Output_min;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidGetProcessVariable(uint16_t DeviceNo, uint32_t PidNo, double *ProcessVariable)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	Pid_ecat_double_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_CONTROL_OUTPUT_MODULE,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(Pid_ecat_double_t));
	if (ret == 0) {
		*ProcessVariable = RetData.doubledata;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidGetSimulateFeedback(uint16_t DeviceNo, uint32_t PidNo, double *Feedback)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	ecat_pid_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_CONTROL_OUTPUT_MODULE,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(ecat_pid_t));
	if (ret == 0) {
		*Feedback = RetData.Feedback;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidGetSetPointValue(uint16_t DeviceNo, uint32_t PidNo, double *SetPointValue)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	Pid_ecat_double_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_SET_POINT_VALUE,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(Pid_ecat_double_t));
	if (ret == 0) {
		*SetPointValue = RetData.doubledata;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_PidGet_Sp_Err_Op_Pv(uint16_t DeviceNo, uint32_t PidNo, double *SetPointValue, double *Error, double *OutputValue, double *ProcessVariable)
{
	int32_t ret;
	Pid_ecat_32_t Ops;
	Pid_ecat_in_out_t RetData;
	Ops.PidNo = PidNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PID_IN_OUT_VALUE,
							&Ops, sizeof(Pid_ecat_32_t), &RetData, sizeof(Pid_ecat_in_out_t));
	if (ret == 0) {
		*SetPointValue = RetData.SetPointValue;
		*Error = RetData.Error;
		*OutputValue = RetData.OutPutValue;
		*ProcessVariable = RetData.ProcessVariable;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisVelocityFeedForwardGain(uint16_t DeviceNo, uint16_t AxisNo, double Gain)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Gain;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_FEEDFORWARD_GAIN,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisVelocityFeedForwardGain(uint16_t DeviceNo, uint16_t AxisNo, double *Gain)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_FEEDFORWARD_GAIN,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0) {
		*Gain = RetData.DoubleValue[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McStewartPlatformMoveAbs_PT(uint16_t DeviceNo, uint16_t GroupNo, double Pose[], double *Pos, double Time)
{
	int32_t ret;
	ecat_stewart_Pose_t Ops;
	ecat_stewart_return_t RetData;
	Ops.GroupNo = GroupNo;
	Ops.time = Time;
	memcpy(Ops.Pose, Pose, sizeof(Ops.Pose));
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_STEWART_PLATFORM_MOVE_ABS_PT,
							&Ops, sizeof(ecat_stewart_Pose_t), &RetData, sizeof(ecat_stewart_return_t));
	if (ret == 0) {
		memcpy(Pos, RetData.Pos, sizeof(RetData.Pos));
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetStewartPlatform_M2(uint16_t DeviceNo, double Bx[], double By[], double Px[], double Py[], double Z0, double RodLength[], double Max_RodLength[])
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	ecat_stewart_t Ops;
	ecat_slave_reg_return_t RetData;
	uint32_t OpsSize = sizeof(ecat_stewart_t);
	uint32_t RetSize = sizeof(ecat_slave_reg_return_t);
	int i, j, error = 0;
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	stewartPlatform_t *Sp = &EcatDevice[DeviceNo].StewartPlatform;

	int32_t ret = 0;
	Sp->Z0 = Z0;
	/******************************/
	for (i = 1; i < 7; i++)
	{
		Sp->Px[i] = Px[i - 1];
		Sp->Py[i] = Py[i - 1];
		Sp->Pz[i] = Z0;

		Sp->Bx[i] = Bx[i - 1];
		Sp->By[i] = By[i - 1];
		Sp->Bz[i] = 0;
	}


	double L[6], dL[6];
	double ca, cb, cr, sa, sb, sr;
	double r11, r12, r13, r21, r22, r23, r31, r32, r33;
	double Xbar[6], Ybar[6], Zbar[6];
	double U[6], V[6], W[6];
	double x, y, z, a, b, r;

	x = 0;
	y = 0;
	z = Sp->Z0;
	a = 0;
	b = 0;
	r = 0;


	ca = cos(a);
	cb = cos(b);
	cr = cos(r);
	sa = sin(a);
	sb = sin(b);
	sr = sin(r);

	r11 = cb*cr;
	r12 = sa*sb*cr - ca*sr;
	r13 = ca*sb*cr + sa*sr;
	r21 = cb*sr;
	r22 = sa*sb*sr + ca*cr;
	r23 = ca*sb*sr - sa*cr;
	r31 = -sb;
	r32 = sa*cb;
	r33 = ca*cb;

	for (i = 0; i<6; i++)
	{
		Xbar[i] = x - Sp->Bx[i + 1];
		Ybar[i] = y - Sp->By[i + 1];
		Zbar[i] = z;
		U[i] = r11*Sp->Px[i + 1] + r12*Sp->Py[i + 1];
		V[i] = r21*Sp->Px[i + 1] + r22*Sp->Py[i + 1];
		W[i] = r31*Sp->Px[i + 1] + r32*Sp->Py[i + 1];
	}


	for (i = 0; i<6; i++)
	{
		L[i] = sqrt((Xbar[i] + U[i]) * (Xbar[i] + U[i]) + (Ybar[i] + V[i]) * (Ybar[i] + V[i]) + (Zbar[i] + W[i]) * (Zbar[i] + W[i]));
		Sp->L0[i] = L[i];
		Sp->Lmin[i] = L[i];
		Sp->Lmax[i] = Max_RodLength[i];

	}

	for (i = 0; i<6; i++)
	{
		Ops.Bx[i] = Sp->Bx[i + 1];
		Ops.By[i] = Sp->By[i + 1];
		Ops.Px[i] = Sp->Px[i + 1];
		Ops.Py[i] = Sp->Py[i + 1];
	}
	Ops.Z0 = Sp->Z0;
	memcpy(Ops.Max_RodLength, Sp->Lmax, sizeof(Sp->Lmax));
	memcpy(Ops.Min_RodLength, Sp->Lmin, sizeof(Sp->Lmin));
	if ((ret = SetCommandData(&EcatDevice[DeviceNo], ECAT_IPC_CMD_SET_STEWART_PLATFORM, (uint32_t*)&Ops, OpsSize)) < 0)
		goto out_return;

	ret = GetReturnData(&EcatDevice[DeviceNo], NULL, 0);

out_return:
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	if (ret == 0)
		Sp->setup = 1;

	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetStewartPlatform_M1(uint16_t DeviceNo, double *radiusB, double *angleB, double *radiusP, double *angleP, double *RodLength, double *Max_RodLength)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	stewartPlatform_t *Sp = &EcatDevice[DeviceNo].StewartPlatform;

	if (Sp->setup != 1)
		return ECAT_ERR_MC_NOT_SETUP;
	*radiusB = Sp->rb;
	*radiusP = Sp->rp;
	*angleB = Sp->ab_degree;
	*angleP = Sp->ap_degree;
	*Max_RodLength = Sp->Lmax[0];
	*RodLength = Sp->L0[0];

	return 0;
}
/*****************************************************************************/
int32_t ECAT_McGetStewartPlatform_M2(uint16_t DeviceNo, double *Bx, double *By, double *Px, double *Py, double *Z0, double *RodLength, double *Max_RodLength)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	stewartPlatform_t *Sp = &EcatDevice[DeviceNo].StewartPlatform;

	if (Sp->setup != 1)
		return ECAT_ERR_MC_NOT_SETUP;
	int i;
	for (i = 0; i<6; i++)
	{
		Bx[i] = Sp->Bx[i + 1];
		By[i] = Sp->By[i + 1];
		Px[i] = Sp->Px[i + 1];
		Py[i] = Sp->Py[i + 1];
		Max_RodLength[i] = Sp->Lmax[i];
		RodLength[i] = Sp->L0[i];
	}
	*Z0 = Sp->Z0;
	return 0;
}
/*****************************************************************************/
int32_t ECAT_McSetGroupVelLimitStatus(uint16_t DeviceNo, uint16_t GroupNo, uint16_t Status)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = Status;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_VEL_LIMIT_STATUS,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetGroupVelLimitStatus(uint16_t DeviceNo, uint16_t GroupNo, uint16_t *Status)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = GroupNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_GROUP_VEL_LIMIT_STATUS,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0) {
		*Status = RetData.U16Value;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetGroupVelLimitValue(uint16_t DeviceNo, uint16_t GroupNo, double Value)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.DoubleData[0] = Value;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_VEL_LIMIT_VALUE,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetGroupVelLimitValue(uint16_t DeviceNo, uint16_t GroupNo, double *Value)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = GroupNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_GROUP_VEL_LIMIT_VALUE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0) {
		*Value = RetData.DoubleValue[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetMotionRecordParam(uint16_t DeviceNo, uint16_t Value1, uint16_t Value2)
{
	ecat_mc_axis_ops_t Ops;
	Ops.U16Data[0] = Value1;
	Ops.U16Data[1] = Value2;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_RECORD_PARAM,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetMotionRecordParam(uint16_t DeviceNo, uint16_t *Value1, uint16_t *Value2)
{
	int32_t ret;
	ecat_mc_axis_ops_t RetData;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_RECORD_PARAM,
							NULL, 0, &RetData, sizeof(ecat_mc_axis_ops_t));
	if (ret == 0) {
		*Value1 = RetData.U16Data[0];
		*Value2 = RetData.U16Data[1];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetMotionRecordValue(uint16_t DeviceNo, uint32_t CountNo, uint16_t AxisNo, float *Value1, float *Value2)
{
	int32_t ret;
	ecat_mc_axis_ops_t Ops;
	ecat_mc_return_float_multi_t RetData;
	uint32_t i;
	Ops.AxisNo = AxisNo;
	Ops.U32Data = CountNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_RECORD_VALUE,
							&Ops, sizeof(ecat_mc_axis_ops_t), &RetData, sizeof(ecat_mc_return_float_multi_t));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++) {
				*(Value1 + i) = RetData.FloatValue_multi_Value1[i];
				*(Value2 + i) = RetData.FloatValue_multi_Value2[i];
			}
		} else 	{
			*Value1 = RetData.FloatValue_multi_Value1[0];
			*Value2 = RetData.FloatValue_multi_Value2[0];
		}
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McClearMotionRecord(uint16_t DeviceNo)
{

	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_CLEAR_RECORD,
							NULL, 0, NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetAxisPosSoftwareLimit(uint16_t DeviceNo, uint16_t AxisNo, double Maximum, double Minimum)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Minimum;
	Ops.DoubleData[1] = Maximum;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_SOFTWARE_LIMIT,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisPosSoftwareLimit(uint16_t DeviceNo, uint16_t AxisNo, double *Maximum, double *Minimum)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_SOFTWARE_LIMIT,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0) {
		*Minimum = RetData.DoubleValue[0];
		*Maximum = RetData.DoubleValue[1];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetProcessTime(uint16_t DeviceNo, double *Time)
{
	int32_t ret;
	ecat_double_t RetData;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PROCESS_TIME,
							NULL, 0, &RetData, sizeof(ecat_double_t));
	if (ret == 0) {
		*Time = RetData.doubledata;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetGroupHold(uint16_t DeviceNo, uint16_t GroupNo, uint16_t Status)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = Status;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_HOLD,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetAxisPosSoftwareLimitStatus(uint16_t DeviceNo, uint16_t AxisNo, uint16_t Status, uint16_t ErrorStop)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = Status;
	Ops.U16Data[1] = ErrorStop;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_POS_SW_STATUS,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisPosSoftwareLimitStatus(uint16_t DeviceNo, uint16_t AxisNo, uint16_t *Status, uint16_t *ErrorStop)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_axis_ops_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_POS_SW_STATUS,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_axis_ops_t));
	if (ret == 0) {
		*Status = RetData.U16Data[0];
		*ErrorStop = RetData.U16Data[1];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisHomeState(uint16_t DeviceNo, uint16_t AxisNo, uint16_t *State)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_HOME_STATE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0) {
		*State = RetData.U16Value;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McAxisGantryIn(uint16_t DeviceNo, uint16_t MasterNo, uint16_t SlaveNo, int32_t Direction)
{
	ecat_mc_axis_ops_t Ops;
	Ops.MasterNo = MasterNo;
	Ops.AxisNo = SlaveNo;
	Ops.I32Data = Direction;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_GANTRY_IN,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetGroupCmdModeEx(uint16_t DeviceNo, uint16_t GroupNo, uint16_t CmdMode)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = CmdMode;
	Ops.U16Data[1] = 1; //put this action(mode changing) in the buffer
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_CMD_MODE,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisGantryMaxPosDiff(uint16_t DeviceNo, uint16_t SlaveNo, double Value)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = SlaveNo;
	Ops.DoubleData[0] = Value;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_GANTRY_MAX_POS_DIFF,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisGantryMaxPosDiffStatus(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t Status)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = SlaveNo;
	Ops.U16Data[0] = Status;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_GANTRY_MAX_POS_DIFF_STATUS,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_EvEnableEvent(uint16_t DeviceNo, uint16_t EventID)
{
	ecat_event_ops_t Ops;
	Ops.EventId = EventID;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_EV_ENABLE_EVENT,
							&Ops, sizeof(ecat_event_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_EvDisableEvent(uint16_t DeviceNo, uint16_t EventID)
{
	ecat_event_ops_t Ops;
	Ops.EventId = EventID;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_EV_DISABLE_EVENT,
							&Ops, sizeof(ecat_event_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_WaitforEvent(uint16_t DeviceNo, uint32_t TimeOut, uint32_t *TriggeredEvent)
{
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	int ret = 0;
	uint32_t event_state;
	struct timespec ts_now, ts_timeout;
	int64_t timeout_ns;

	pthread_mutex_lock(&EcatDevice[DeviceNo].cond_event_mutex);

	if (TimeOut == 0xFFFFFFFF) {
		if (pthread_cond_wait(&EcatDevice[DeviceNo].cond_event, &EcatDevice[DeviceNo].cond_event_mutex) != 0)
			ret = ECAT_ERR_DRV_IOCTL;
	} else {
		clock_gettime(CLOCK_REALTIME, &ts_now);
		timeout_ns = ((int64_t)TimeOut * 1000000);
		timespec_add_nsec(&ts_timeout, &ts_now, timeout_ns);
		if ((ret = pthread_cond_timedwait(&EcatDevice[DeviceNo].cond_event,
				&EcatDevice[DeviceNo].cond_event_mutex, &ts_timeout)) != 0) {
			if (ret == ETIMEDOUT)
				ret = ECAT_ERR_IPC_TIME_OUT;
			else
				ret = ECAT_ERR_DRV_IOCTL;
		}
	}

	pthread_mutex_unlock(&EcatDevice[DeviceNo].cond_event_mutex);
	if(ret == 0) {
		SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
		read_dpram(EcatDevice[DeviceNo].Fd, 0x7FD, &event_state, 1);
		SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
		*TriggeredEvent = event_state;
		return 0;
	}else
		return ret;
}
/*****************************************************************************/
int32_t ECAT_EvSetComparePositionParameters(uint16_t DeviceNo, uint16_t EventID, uint16_t AxisNo, uint16_t Operator, double ComparePosition)
{
	ecat_event_ops_t Ops;
	Ops.EventId = EventID;
	Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = Operator;
	Ops.DoubleData[0] = ComparePosition;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_EV_SET_COMPARE_POSITION_PARAMETERS,
							&Ops, sizeof(ecat_event_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_EvSetCompareDIBitParameters(uint16_t DeviceNo, uint16_t EventID, uint16_t SlaveNo, uint16_t BitNo, uint32_t CompareValue)
{
	ecat_event_ops_t Ops;
	Ops.EventId = EventID;
	Ops.SlaveNo = SlaveNo;
	Ops.U16Data[0] = BitNo;
	Ops.U32Data[0] = CompareValue;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_EV_SET_COMPARE_DIBIT_PARAMETERS,
							&Ops, sizeof(ecat_event_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_EvSetMotionCompleteParameters(uint16_t DeviceNo, uint16_t EventID, uint16_t AxisNo)
{
	ecat_event_ops_t Ops;
	Ops.EventId = EventID;
	Ops.AxisNo = AxisNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_EV_SET_MOTION_COMPLETE_PARAMETERS,
							&Ops, sizeof(ecat_event_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveAO(uint16_t DeviceNo, uint16_t GroupNo, uint16_t SlaveNo, uint32_t RunMode, uint16_t ChannelNo, uint16_t RawData, double VoltData)
{
	ecat_mc_group_move_ao_ops_t Ops;
	Ops.RunMode = RunMode;
	Ops.GroupNo = GroupNo;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	Ops.RawData = RawData;
	Ops.VoltData = VoltData;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GROUP_MOVE_AO,
							&Ops, sizeof(ecat_mc_group_move_ao_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetGroupAccTimeEx(uint16_t DeviceNo, uint16_t GroupNo, uint16_t Time_ms)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = Time_ms;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_ACC_TIME_EX,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetGroupBlendingPercent(uint16_t DeviceNo, uint16_t GroupNo, uint16_t Value)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = Value;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_BLENDING_PERCENT,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetProfileInterval(uint16_t DeviceNo, uint16_t ProfileNo, uint16_t Value)
{
	ecat_mc_axis_ops_t Ops;
	Ops.U16Data[0] = ProfileNo;
	Ops.U16Data[1] = Value;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_PROFILE_INTERVAl,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveSuperimposed(uint16_t DeviceNo, uint16_t AxisNo, double Pos, double Vel)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Pos;
	Ops.Vel = Vel;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_MOVE_SUPERIMPOSED,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisHaltSuperimposed(uint16_t DeviceNo, uint16_t AxisNo)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_HALT_SUPERIMPOSED,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetMotionRecordValueEx(uint16_t DeviceNo, uint32_t CountNo, uint16_t Count, uint16_t AxisNo, float *Value1, float *Value2, uint16_t *ActualCount)
{
	int32_t ret;
	uint16_t i;
	ecat_mc_axis_ops_t Ops;
	ecat_mc_return_float_multi_t RetData;
	Ops.AxisNo = AxisNo;
	Ops.U32Data = CountNo;
	Ops.U16Data[0] = Count;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_RECORD_VALUE_EX,
							&Ops, sizeof(ecat_mc_axis_ops_t), &RetData, sizeof(ecat_mc_return_float_multi_t));
	if (ret == 0) {
		*ActualCount = RetData.ActualCount;
		for (i = 0; i < *ActualCount; i++) {
			*(Value1 + i) = RetData.FloatValue_multi_Value1[i];
			*(Value2 + i) = RetData.FloatValue_multi_Value2[i];
		}
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_SetHeartBeat(uint16_t DeviceNo, uint32_t Value)
{
	ecat_u32_t Ops;
	Ops.u32data = Value;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_HEARTBEAT,
							&Ops, sizeof(ecat_u32_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_SetHeartBeatStatus(uint16_t DeviceNo, uint32_t Value)
{
	ecat_u32_t Ops;
	Ops.u32data = Value;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_HEARTBEAT_STATUS,
							&Ops, sizeof(ecat_u32_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetGroupBlendingPercentEx(uint16_t DeviceNo, uint16_t GroupNo, uint16_t Value)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = Value;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_BLENDING_PERCENT_EX,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_EvSetCompareDIParameters(uint16_t DeviceNo, uint16_t EventID, uint16_t SlaveNo, uint16_t OffsetByte, uint32_t CompareValue, uint32_t Mask)
{
	ecat_event_ops_t Ops;
	Ops.EventId = EventID;
	Ops.SlaveNo = SlaveNo;
	Ops.U16Data[0] = OffsetByte;
	Ops.U32Data[0] = CompareValue;
	Ops.U32Data[1] = Mask;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_EV_SET_COMPARE_DI_PARAMETERS,
							&Ops, sizeof(ecat_event_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetGroupPause(uint16_t DeviceNo, uint16_t GroupNo, uint16_t Status)
{
	ecat_mc_group_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.U16Data[0] = Status;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_GROUP_PAUSE,
							&Ops, sizeof(ecat_mc_group_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_EvSetCompareAxisStateParameters(uint16_t DeviceNo, uint16_t EventID, uint16_t AxisNo, uint32_t CompareState)
{
	ecat_event_ops_t Ops;
	Ops.EventId = EventID;
	Ops.AxisNo = AxisNo;
	Ops.U32Data[0] = CompareState;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_EV_SET_COMPARE_AXIS_STATE_PARAMETERS,
							&Ops, sizeof(ecat_event_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_SetSlaveAiProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, uint8_t Range)
{
	int32_t ret;
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_SET_AI_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	Ops.Range = Range;
	ret = CommonSlaveAIO(DeviceNo, &Ops);
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveAiProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, uint8_t *Range)
{
	int32_t ret;
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_GET_AI_PROPERTY;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	ret = CommonSlaveAIO(DeviceNo, &Ops);
	if (ret == 0)
		*Range = Ops.Range;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveAiRawData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, int16_t *Data)
{
	int32_t ret;
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_GET_AI_RAW_DATA;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	ret = CommonSlaveAIO(DeviceNo, &Ops);
	if (ret == 0)
		*Data = Ops.RawData;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveAiVoltData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, double *Data)
{
	int32_t ret;
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_GET_AI_VOLT_DATA;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	ret = CommonSlaveAIO(DeviceNo, &Ops);
	if (ret == 0)
		*Data = Ops.VoltData;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_GetSlaveAimAData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, double *Data)
{
	int32_t ret;
	ecat_slave_aio_data_t Ops;
	Ops.RunMode = SLAVE_AIO_RM_GET_AI_mA_DATA;
	Ops.SlaveNo = SlaveNo;
	Ops.ChannelNo = ChannelNo;
	ret = CommonSlaveAIO(DeviceNo, &Ops);
	if (ret == 0)
		*Data = Ops.mAData;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McAxisHomeEx(uint16_t DeviceNo, uint16_t AxisNo, uint16_t Settings)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
    Ops.U16Data[0] = Settings;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_HOME_EX,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_EvSetCompareCmdPositionParameters(uint16_t DeviceNo, uint16_t EventID, uint16_t AxisNo, uint16_t Operator, double ComparePosition)
{
	ecat_event_ops_t Ops;
	Ops.EventId = EventID;
	Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = Operator;
	Ops.DoubleData[0] = ComparePosition;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_EV_SET_COMPARE_COMMAND_POSITION_PARAMETERS,
							&Ops, sizeof(ecat_event_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_EvSetCompareAxisVelStateParameters(uint16_t DeviceNo, uint16_t EventID, uint16_t AxisNo, uint32_t CompareState)
{
	ecat_event_ops_t Ops;
	Ops.EventId = EventID;
	Ops.AxisNo = AxisNo;
    Ops.U32Data[0] = CompareState;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_EV_SET_COMPARE_AXIS_VEL_STATE_PARAMETERS,
							&Ops, sizeof(ecat_event_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetAxisMaxVelocity(uint16_t DeviceNo, uint16_t AxisNo, double MaxVelocity)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = MaxVelocity;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_MAX_VELOCITY,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisMaxVelocity(uint16_t DeviceNo, uint16_t AxisNo, double *MaxVelocity)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_MAX_VELOCITY,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0) {
		*MaxVelocity = RetData.DoubleValue[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveLineAbs_P2P(uint16_t DeviceNo, uint16_t GroupNo, double Pos[])
{
	return SendMcGroupMoveLine(DeviceNo, GroupNo, MC_LINE_MODE_P2P, true, 0, 0, Pos, NULL);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveLineRel_P2P(uint16_t DeviceNo, uint16_t GroupNo, double Pos[])
{
	return SendMcGroupMoveLine(DeviceNo, GroupNo, MC_LINE_MODE_P2P, false, 0, 0, Pos, NULL);
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveAbs_P2P(uint16_t DeviceNo, uint16_t AxisNo, double Pos)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Pos;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_MOVE_ABS_P2P,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveRel_P2P(uint16_t DeviceNo, uint16_t AxisNo, double Pos)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Pos;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_MOVE_REL_P2P,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
void Cal_ECAT2016_Scale(ecat_slave_aio_config_t *AioConfig)
{
    switch (AioConfig->Range) {
    case ECAT2016_AI_BI_10V:
        AioConfig->VoltScaling = 10.0 / 0x7FFF;
        break;
    case ECAT2016_AI_BI_5V:
        AioConfig->VoltScaling = 5.0 / 0x7FFF;
        break;
    case ECAT2016_AI_BI_2_5V:
        AioConfig->VoltScaling = 2.5 / 0x7FFF;
        break;
    case ECAT2016_AI_BI_1V:
        AioConfig->VoltScaling = 1.0 / 0x7FFF;
        break;
    case ECAT2016_AI_BI_500mV:
        AioConfig->VoltScaling = 0.5 / 0x7FFF;
        break;
    case ECAT2016_AI_BI_100mV:
        AioConfig->VoltScaling = 0.1 / 0x7FFF;
        break;
    case ECAT2016_AI_BI_50mV:
        AioConfig->VoltScaling = 0.05 / 0x7FFF;
        break;
    case ECAT2016_AI_BI_25mV:
        AioConfig->VoltScaling = 0.025 / 0x7FFF;
        break;
    case ECAT2016_AI_BI_20mV:
        AioConfig->VoltScaling = 0.02 / 0x7FFF;
        break;
    case ECAT2016_AI_BI_16mV:
        AioConfig->VoltScaling = 0.016 / 0x7FFF;
        break;
    case ECAT2016_AI_BI_15mV:
        AioConfig->VoltScaling = 0.015 / 0x7FFF;
        break;
    }
}
/*****************************************************************************/
int32_t ECAT_Set_ECAT2016_AiProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, uint8_t Range, uint32_t *AbortCode)
{
    CHECK_ECAT_SLAVE_NO(SlaveNo);
    int32_t ret;
    uint16_t Index;
	if (Range > ECAT2016_AI_BI_15mV)
		return ECAT_ERR_INVALID_PARAM;
    ecat_slave_aio_config_t *AioConfig = &EcatDevice[DeviceNo].Slave[SlaveNo].AioConfig;
	Index = 0x8000 + (0x10 * ChannelNo);
    ret = ECAT_SetSlaveSdoObject(DeviceNo, SlaveNo, Index, 0x19, 4, Range, AbortCode);
    if(ret == 0)
    {
        AioConfig->Range = Range;
        Cal_ECAT2016_Scale(AioConfig);
    }
    return ret;
}
/*****************************************************************************/
int32_t ECAT_Get_ECAT2016_AiProperty(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, uint8_t *Range, uint32_t *AbortCode)
{
    CHECK_ECAT_SLAVE_NO(SlaveNo);
	int32_t ret;
	uint16_t Index;
	uint32_t ObjectVal;

	Index = 0x8000 + (0x10 * ChannelNo);
	ret = ECAT_GetSlaveSdoObject(DeviceNo, SlaveNo, Index, 0x19, 4, &ObjectVal, AbortCode);
    if(ret == 0)
    {
        ecat_slave_aio_config_t *AioConfig = &EcatDevice[DeviceNo].Slave[SlaveNo].AioConfig;
        AioConfig->Range = ObjectVal;
        Cal_ECAT2016_Scale(AioConfig);
        *Range = ObjectVal;
    }
    return ret;
}
/*****************************************************************************/
int32_t ECAT_Get_ECAT2016_AiRawData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, int16_t *Data)
{
  	int32_t ret;
    ecat_2016_t Data_st;
    ret = ECAT_GetSlaveTxPdoData(DeviceNo,  SlaveNo, sizeof(ecat_2016_t) * ChannelNo, sizeof(ecat_2016_t), (uint8_t *)&Data_st);
    if(ret == 0)
        *Data = Data_st.Value;
    else if(ret == ECAT_ERR_INVALID_OFFSET)
        ret = ECAT_ERR_INVALID_CHANNEL;
    return ret;
}
/*****************************************************************************/
int32_t ECAT_Get_ECAT2016_AiVoltData(uint16_t DeviceNo, uint16_t SlaveNo, uint16_t ChannelNo, double *Data)
{
    CHECK_ECAT_SLAVE_NO(SlaveNo);
  	int32_t ret;
    ecat_2016_t Data_st;
    ret = ECAT_GetSlaveTxPdoData(DeviceNo,  SlaveNo, sizeof(ecat_2016_t) * ChannelNo, sizeof(ecat_2016_t), (uint8_t *)&Data_st);
    if(ret == 0)
    {
        ecat_slave_aio_config_t *AioConfig = &EcatDevice[DeviceNo].Slave[SlaveNo].AioConfig;
        *Data = Data_st.Value *AioConfig->VoltScaling;
    }
    else if(ret == ECAT_ERR_INVALID_OFFSET)
        ret = ECAT_ERR_INVALID_CHANNEL;
    return ret;
}
/*****************************************************************************/
int32_t ECAT_EvSetCompareAiParameters(uint16_t DeviceNo, uint16_t EventID, uint16_t Operator, uint16_t SlaveNo, uint16_t OffsetByte, uint16_t DataSize, double ScaleGain, double ScaleOffset, double CompareValue)
{
	ecat_event_ops_t Ops;
	Ops.EventId = EventID;
	Ops.SlaveNo= SlaveNo;
	Ops.U16Data[0] = Operator;
	Ops.U16Data[1] = OffsetByte;
	Ops.U32Data[0] = DataSize;
	Ops.DoubleData[0] = ScaleGain;
	Ops.DoubleData[1] = ScaleOffset;
	Ops.DoubleData[2] = CompareValue;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_EV_SET_COMPARE_AI_PARAMETERS,
							&Ops, sizeof(ecat_event_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McSetAxisTorqueGain(uint16_t DeviceNo, uint16_t AxisNo, double Gain, uint32_t *AbortCode)
{
	int32_t ret;
	ecat_mc_axis_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Gain;
	ret = GenericSendCommand(DeviceNo, 1000, ECAT_IPC_CMD_MC_SET_AXIS_TORQUE_GAIN,
							&Ops, sizeof(ecat_mc_axis_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret != 0) {
		*AbortCode = RetData.U32Value;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisTorqueGain(uint16_t DeviceNo, uint16_t AxisNo, double *Gain)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_TORQUE_GAIN,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0) {
		*Gain = RetData.DoubleValue[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisRatedTorque(uint16_t DeviceNo, uint16_t AxisNo, double *Torque, uint32_t *AbortCode)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_RATED_TORQUE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
        *Torque = RetData.DoubleValue[0];
    else
		*AbortCode = RetData.U32Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveTor(uint16_t DeviceNo, uint16_t AxisNo, double Torque)
{
	ecat_mc_axis_ops_t Ops;
	Ops.AxisNo = AxisNo;
	Ops.DoubleData[0] = Torque;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_AXIS_MOVE_TOR,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisActualTorque(uint16_t DeviceNo, uint16_t AxisNo, double *Torque)
{
	int32_t ret;
	uint16_t i;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_double_multi_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_ACTUAL_TORQUE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_double_multi_t));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++)
				*(Torque + i) = RetData.DoubleValue_multi[i];
		} else
			*Torque = RetData.DoubleValue_multi[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisMaxTorque(uint16_t DeviceNo, uint16_t AxisNo, double Torque, uint32_t *AbortCode)
{
	int32_t ret;
	ecat_mc_axis_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.AxisNo = AxisNo;
    Ops.DoubleData[0] = Torque;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_MAX_TORQUE,
							&Ops, sizeof(ecat_mc_axis_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret != 0) {
		*AbortCode = RetData.U32Value;
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisMaxTorque(uint16_t DeviceNo, uint16_t AxisNo, double *Torque, uint32_t *AbortCode)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_MAX_TORQUE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0)
        *Torque = RetData.DoubleValue[0];
    else
		*AbortCode = RetData.U32Value;
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McGetAxisCommandVel(uint16_t DeviceNo, uint16_t AxisNo, double *Vel)
{
	int32_t ret;
	uint16_t i;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_double_multi_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_CMD_VEL,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_double_multi_t));
	if (ret == 0) {
		if (AxisNo == 65535) {
			for (i = 0; i < MC_AXIS_NO_MAX; i++)
				*(Vel + i) = RetData.DoubleValue_multi[i];
		} else
			*Vel = RetData.DoubleValue_multi[0];
	}
	return ret;
}
/*****************************************************************************/
int32_t SendMovePathCtrlAction(uint16_t DeviceNo, uint8_t Read, ecat_mc_move_path_ops_t *Ops)
{
	if (Read) {
		return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_MOVE_PATH_CTRL_ACTION,
								Ops, sizeof(ecat_mc_move_path_ops_t), Ops, sizeof(ecat_mc_move_path_ops_t));
	} else {
		return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_MOVE_PATH_CTRL_ACTION,
								Ops, sizeof(ecat_mc_move_path_ops_t), NULL, 0);
	}
}
/*****************************************************************************/
int32_t ECAT_McGroupMovePath(uint16_t DeviceNo, uint16_t GroupNo, uint16_t PathDataNo, uint8_t Restart, uint16_t DataIndex, uint8_t Repeat)
{
	ecat_mc_move_path_ops_t Ops;
	Ops.GroupNo = GroupNo;
	Ops.PathDataNo = PathDataNo;
	Ops.DataIndex = DataIndex;
	Ops.Restart = Restart;
	Ops.Repeat = Repeat;
	Ops.ActionType = MC_MPCA_MOVING;
	return SendMovePathCtrlAction(DeviceNo, 0, &Ops);
}
/*****************************************************************************/
int32_t ECAT_McAddPathData(uint16_t DeviceNo, uint16_t PathDataNo, uint32_t CmdType,
		uint8_t AbsMove, double EndPos[], double AuxPos[], double Args[])
{
	ecat_mc_move_path_ops_t Ops;
	Ops.PathDataNo = PathDataNo;
	Ops.Queue.CmdType = CmdType;
	Ops.Queue.AbsMove = AbsMove;
	Ops.ActionType = MC_MPCA_ADD_DATA;
	if (EndPos)
		memcpy(Ops.Queue.EndPos, EndPos, sizeof(Ops.Queue.EndPos));
	if (AuxPos)
		memcpy(Ops.Queue.AuxPos, AuxPos, sizeof(Ops.Queue.AuxPos));
	if (Args)
		memcpy(Ops.Queue.Args, Args, sizeof(Ops.Queue.Args));
	return SendMovePathCtrlAction(DeviceNo, 0, &Ops);
}
/*****************************************************************************/
int32_t ECAT_McSetPathData(uint16_t DeviceNo, uint16_t PathDataNo, uint16_t DataIndex, uint32_t CmdType,
		uint8_t AbsMove, double EndPos[], double AuxPos[], double Args[])
{
	ecat_mc_move_path_ops_t Ops;
	Ops.PathDataNo = PathDataNo;
	Ops.DataIndex = DataIndex;
	Ops.Queue.CmdType = CmdType;
	Ops.Queue.AbsMove = AbsMove;
	Ops.ActionType = MC_MPCA_SET_DATA;
	if (EndPos)
		memcpy(Ops.Queue.EndPos, EndPos, sizeof(Ops.Queue.EndPos));
	if (AuxPos)
		memcpy(Ops.Queue.AuxPos, AuxPos, sizeof(Ops.Queue.AuxPos));
	if (Args)
		memcpy(Ops.Queue.Args, Args, sizeof(Ops.Queue.Args));
	return SendMovePathCtrlAction(DeviceNo, 0, &Ops);
}
/*****************************************************************************/
int32_t ECAT_McGetPathData(uint16_t DeviceNo, uint16_t PathDataNo, uint16_t DataIndex, uint32_t *CmdType,
		uint8_t *AbsMove, double EndPos[], double AuxPos[], double Args[])
{
	ecat_mc_move_path_ops_t Ops;
	int32_t ret;
	Ops.PathDataNo = PathDataNo;
	Ops.DataIndex = DataIndex;
	Ops.ActionType = MC_MPCA_GET_DATA;
	ret = SendMovePathCtrlAction(DeviceNo, 1, &Ops);
	if (ret == 0) {
		*CmdType = Ops.Queue.CmdType;
		*AbsMove = Ops.Queue.AbsMove;
		if (EndPos)
			memcpy(EndPos, Ops.Queue.EndPos, sizeof(Ops.Queue.EndPos));
		if (AuxPos)
			memcpy(AuxPos, Ops.Queue.AuxPos, sizeof(Ops.Queue.AuxPos));
		if (Args)
			memcpy(Args, Ops.Queue.Args, sizeof(Ops.Queue.Args));
	}

	return ret;
}
/*****************************************************************************/
int32_t ECAT_McClearPathData(uint16_t DeviceNo, uint16_t PathDataNo)
{
	ecat_mc_move_path_ops_t Ops;
	Ops.PathDataNo = PathDataNo;
	Ops.ActionType = MC_MPCA_CLEAR_DATA;
	return SendMovePathCtrlAction(DeviceNo, 0, &Ops);
}
/*****************************************************************************/
int32_t ECAT_McGetPathDataSize(uint16_t DeviceNo, uint16_t PathDataNo, uint16_t *Size)
{
	ecat_mc_move_path_ops_t Ops;
	int32_t ret;
	Ops.PathDataNo = PathDataNo;
	Ops.ActionType = MC_MPCA_GET_DATA_SIZE;
	ret = SendMovePathCtrlAction(DeviceNo, 1, &Ops);
	if (ret == 0)
		*Size = Ops.DataIndex;

	return ret;
}
/*****************************************************************************/
int32_t SendMcMoveAdv(uint16_t DeviceNo, ecat_mc_move_adv_t *Ops, uint8_t AccDecMode)
{
	if (AccDecMode >= MC_ACC_DEC_MODE_TIME) {
		Ops->Accel = fabs((Ops->ReqVel - Ops->StartVel) / Ops->Accel);
		Ops->Decel = fabs((Ops->ReqVel - Ops->FinalVel) / Ops->Decel);
	}

	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_MOVE_ADV,
							Ops, sizeof(ecat_mc_move_adv_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveAbsAdv(uint16_t DeviceNo, uint16_t AxisNo, double EndPos,
		double StartVel, double ReqVel, double FinalVel, double Accel, double Decel, uint8_t AccDecMode)
{
	ecat_mc_move_adv_t Ops;
	Ops.No = AxisNo;
	Ops.GroupMove = 0;
	Ops.AbsMove = 1;
	Ops.StartVel = StartVel;
	Ops.ReqVel = ReqVel;
	Ops.FinalVel = FinalVel;
	Ops.Accel = Accel;
	Ops.Decel = Decel;
	Ops.EndPos[0] = EndPos;
	return SendMcMoveAdv(DeviceNo, &Ops, AccDecMode);
}
/*****************************************************************************/
int32_t ECAT_McAxisMoveRelAdv(uint16_t DeviceNo, uint16_t AxisNo, double EndPos,
		double StartVel, double ReqVel, double FinalVel, double Accel, double Decel, uint8_t AccDecMode)
{
	ecat_mc_move_adv_t Ops;
	Ops.No = AxisNo;
	Ops.GroupMove = 0;
	Ops.AbsMove = 0;
	Ops.StartVel = StartVel;
	Ops.ReqVel = ReqVel;
	Ops.FinalVel = FinalVel;
	Ops.Accel = Accel;
	Ops.Decel = Decel;
	Ops.EndPos[0] = EndPos;
	return SendMcMoveAdv(DeviceNo, &Ops, AccDecMode);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveLineAbsAdv(uint16_t DeviceNo, uint16_t GroupNo, double EndPos[],
		double StartVel, double ReqVel, double FinalVel, double Accel, double Decel, uint8_t AccDecMode)
{
	ecat_mc_move_adv_t Ops;
	Ops.No = GroupNo;
	Ops.GroupMove = 1;
	Ops.AbsMove = 1;
	Ops.StartVel = StartVel;
	Ops.ReqVel = ReqVel;
	Ops.FinalVel = FinalVel;
	Ops.Accel = Accel;
	Ops.Decel = Decel;
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	memcpy(Ops.EndPos, EndPos, sizeof(double)*EcatDevice[DeviceNo].GrpAxisCnt[GroupNo]);
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return SendMcMoveAdv(DeviceNo, &Ops, AccDecMode);
}
/*****************************************************************************/
int32_t ECAT_McGroupMoveLineRelAdv(uint16_t DeviceNo, uint16_t GroupNo, double EndPos[],
		double StartVel, double ReqVel, double FinalVel, double Accel, double Decel, uint8_t AccDecMode)
{
	ecat_mc_move_adv_t Ops;
	Ops.No = GroupNo;
	Ops.GroupMove = 1;
	Ops.AbsMove = 0;
	Ops.StartVel = StartVel;
	Ops.ReqVel = ReqVel;
	Ops.FinalVel = FinalVel;
	Ops.Accel = Accel;
	Ops.Decel = Decel;
	CHECK_ECAT_DEVICE_NO(DeviceNo)
	CHECK_ECAT_DEVICE_IS_OPEN(EcatDevice[DeviceNo].IsOpen)
	SEMEPHORE_P_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	memcpy(Ops.EndPos, EndPos, sizeof(double)*EcatDevice[DeviceNo].GrpAxisCnt[GroupNo]);
	SEMEPHORE_V_OPERATION(EcatDevice[DeviceNo].ThreadSem)
	return SendMcMoveAdv(DeviceNo, &Ops, AccDecMode);
}
/*****************************************************************************/
int32_t ECAT_McAxisMove_CiA402_PP(uint16_t DeviceNo, uint16_t AxisNo, uint8_t Abort, uint8_t AbsMove,
		double EndPos, double Vel, double Accel, double Decel)
{
	ecat_mc_move_cia402_profile_t Ops;
	Ops.No = AxisNo;
	Ops.Mode = CIA402_PROFILE_POSITION;
	Ops.Abort = Abort;
	Ops.AbsMove = AbsMove;
	Ops.Vel = Vel;
	Ops.Accel = Accel;
	Ops.Decel = Decel;
	Ops.EndPos = EndPos;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_MOVE_CIA402_PROFILE_MODE,
							&Ops, sizeof(ecat_mc_move_cia402_profile_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisMove_CiA402_PV(uint16_t DeviceNo, uint16_t AxisNo, double Vel, double Accel, double Decel)
{
	ecat_mc_move_cia402_profile_t Ops;
	Ops.No = AxisNo;
	Ops.Mode = CIA402_PROFILE_VOLOCITY;
	Ops.Vel = Vel;
	Ops.Accel = Accel;
	Ops.Decel = Decel;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_MOVE_CIA402_PROFILE_MODE,
							&Ops, sizeof(ecat_mc_move_cia402_profile_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McAxisMove_CiA402_PT(uint16_t DeviceNo, uint16_t AxisNo, double Torque, double Slope)
{
	ecat_mc_move_cia402_profile_t Ops;
	Ops.No = AxisNo;
	Ops.Mode = CIA402_PROFILE_TORQUE;
	Ops.EndPos = Torque;
	Ops.Accel = Slope;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_MOVE_CIA402_PROFILE_MODE,
							&Ops, sizeof(ecat_mc_move_cia402_profile_t), NULL, 0);
}
/*****************************************************************************/
int32_t	ECAT_SetTxPdoBufParam(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t SlaveNo, uint16_t OffsetByte, uint16_t DataSize, uint16_t Enable)
{
	ecat_pdo_rec_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
    Ops.No = SlaveNo;
    Ops.OffsetByte = OffsetByte;
    Ops.DataSize = DataSize;
    Ops.Enable = Enable;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_TX_BUFFER_PARAM,
							&Ops, sizeof(ecat_pdo_rec_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t	ECAT_GetTxPdoBufParam(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t *SlaveNo, uint16_t *OffsetByte, uint16_t *DataSize, uint16_t *Enable)
{
	int32_t ret;
	ecat_pdo_rec_ops_t Ops;
	Ops.ChannelNo = ChannelNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_TX_BUFFER_PARAM,
							&Ops, sizeof(ecat_pdo_rec_ops_t), &Ops, sizeof(ecat_pdo_rec_ops_t));
	if (ret == 0) {
        *SlaveNo = Ops.No;
        *OffsetByte = Ops.OffsetByte;
        *DataSize = Ops.DataSize;
        *Enable = Ops.Enable;
	}
	return ret;
}
/*****************************************************************************/
int32_t	ECAT_SetTxPdoBufEnable(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t Enable)
{
	ecat_pdo_rec_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
    Ops.Enable = Enable;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_TX_BUFFER_ENABLE,
							&Ops, sizeof(ecat_pdo_rec_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t	ECAT_GetTxPdoBufEnable(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t *Enable)
{
	int32_t ret;
	ecat_pdo_rec_ops_t Ops;
	Ops.ChannelNo = ChannelNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_TX_BUFFER_ENABLE,
							&Ops, sizeof(ecat_pdo_rec_ops_t), &Ops, sizeof(ecat_pdo_rec_ops_t));
	if (ret == 0) {
		*Enable = Ops.Enable;
	}
	return ret;
}
/*****************************************************************************/
int32_t	ECAT_GetTxPdoBufValue(uint16_t DeviceNo, uint16_t ChannelNo, float *Data, uint16_t Size, uint16_t *ActualGetSize)
{
	int32_t ret;
	ecat_pdo_rec_ops_t Ops;
	ecat_pdo_rec_return_t RetData;
	Ops.ChannelNo = ChannelNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_TX_BUFFER_VALUE,
							&Ops, sizeof(ecat_pdo_rec_ops_t), &RetData, sizeof(ecat_pdo_rec_return_t));
	if (ret == 0) {
        *ActualGetSize = RetData.ActualGetCount;
        if(Size > sizeof(RetData.Data))
        	memcpy(Data, RetData.Data, sizeof(RetData.Data));
        else
        	memcpy(Data, RetData.Data, Size);
	}
	return ret;
}
/*****************************************************************************/
int32_t ECAT_McSetAxisHomeTorque(uint16_t DeviceNo, uint16_t AxisNo, uint16_t Torque)
{
	ecat_mc_axis_ops_t Ops;
    Ops.AxisNo = AxisNo;
	Ops.U16Data[0] = Torque;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_SET_AXIS_HOME_TORQUE,
							&Ops, sizeof(ecat_mc_axis_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t ECAT_McGetAxisHomeTorque(uint16_t DeviceNo, uint16_t AxisNo, uint16_t *Torque)
{
	int32_t ret;
	ecat_ipc_base_ops_t Ops;
	ecat_mc_return_t RetData;
	Ops.No = AxisNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_MC_GET_AXIS_HOME_TORQUE,
							&Ops, sizeof(ecat_ipc_base_ops_t), &RetData, sizeof(ecat_mc_return_t));
	if (ret == 0) {
		*Torque = RetData.U16Value;
	}
	return ret;
}
/*****************************************************************************/
int32_t	ECAT_SetAiFilterParam(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t SlaveNo, uint16_t OffsetByte, uint16_t DataSize, uint16_t Enable)
{
	ecat_ai_filter_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
    Ops.No = SlaveNo;
    Ops.OffsetByte = OffsetByte;
    Ops.DataSize = DataSize;
    Ops.Enable = Enable;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_AI_FILTER_PARAM,
							&Ops, sizeof(ecat_ai_filter_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t	ECAT_GetAiFilterParam(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t *SlaveNo, uint16_t *OffsetByte, uint16_t *DataSize, uint16_t *Enable)
{
	int32_t ret;
	ecat_ai_filter_ops_t Ops;
	Ops.ChannelNo = ChannelNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_AI_FILTER_PARAM,
							&Ops, sizeof(ecat_ai_filter_ops_t), &Ops, sizeof(ecat_ai_filter_ops_t));
	if (ret == 0) {
        *SlaveNo = Ops.No;
        *OffsetByte = Ops.OffsetByte;
        *DataSize = Ops.DataSize;
        *Enable = Ops.Enable;
	}
	return ret;
}
/*****************************************************************************/
int32_t	ECAT_SetAiFilterEnable(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t Enable)
{
	ecat_ai_filter_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
    Ops.Enable = Enable;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_AI_FILTER_ENABLE,
							&Ops, sizeof(ecat_ai_filter_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t	ECAT_GetAiFilterEnable(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t *Enable)
{
	int32_t ret;
	ecat_ai_filter_ops_t Ops;
	Ops.ChannelNo = ChannelNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_AI_FILTER_ENABLE,
							&Ops, sizeof(ecat_ai_filter_ops_t), &Ops, sizeof(ecat_ai_filter_ops_t));
	if (ret == 0) {
		*Enable = Ops.Enable;
	}
	return ret;
}
/*****************************************************************************/
int32_t	ECAT_SetAiFilterFreq(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t FilterType, double Frequency)
{
	ecat_ai_filter_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
    Ops.FilterType = FilterType;
    Ops.Frequency = Frequency;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_AI_FILTER_FREQUENCY,
							&Ops, sizeof(ecat_ai_filter_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t	ECAT_GetAiFilterFreq(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t FilterType, double *Frequency)
{
	int32_t ret;
	ecat_ai_filter_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
    Ops.FilterType = FilterType;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_AI_FILTER_FREQUENCY,
							&Ops, sizeof(ecat_ai_filter_ops_t), &Ops, sizeof(ecat_ai_filter_ops_t));
	if (ret == 0) {
		*Frequency = Ops.Frequency;
	}
	return ret;
}
/*****************************************************************************/
int32_t	ECAT_GetAiFilterOutput(uint16_t DeviceNo, uint16_t ChannelNo, int32_t *Output)
{
	int32_t ret;
	ecat_ai_filter_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_AI_FILTER_OUTPUT,
							&Ops, sizeof(ecat_ai_filter_ops_t), &Ops, sizeof(ecat_ai_filter_ops_t));
	if (ret == 0) {
		*Output = Ops.Output;
	}
	return ret;
}
/*****************************************************************************/
int32_t	ECAT_SetPdoInToOutParam(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t SlaveNoIn, uint16_t OffsetByteIn, uint16_t DataSizeIn, uint16_t SlaveNoOut, uint16_t OffsetByteOut, uint16_t DataSizeOut)
{
	ecat_pdo_i2o_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
    Ops.No[0] = SlaveNoIn;
    Ops.OffsetByte[0] = OffsetByteIn;
    Ops.No[1] = SlaveNoOut;
    Ops.OffsetByte[1] = OffsetByteOut;
    Ops.DataSize[0] = DataSizeIn;
    Ops.DataSize[1] = DataSizeOut;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_PDO_I2O_PARAM,
							&Ops, sizeof(ecat_pdo_i2o_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t	ECAT_GetPdoInToOutParam(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t *SlaveNoIn, uint16_t *OffsetByteIn, uint16_t *DataSizeIn, uint16_t *SlaveNoOut, uint16_t *OffsetByteOut, uint16_t *DataSizeOut)
{
	int32_t ret;
	ecat_pdo_i2o_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PDO_I2O_PARAM,
							&Ops, sizeof(ecat_pdo_i2o_ops_t), &Ops, sizeof(ecat_pdo_i2o_ops_t));
	if (ret == 0) {
        *SlaveNoIn = Ops.No[0];
        *OffsetByteIn = Ops.OffsetByte[0];
        *SlaveNoOut = Ops.No[1];
        *OffsetByteOut = Ops.OffsetByte[1];
        *DataSizeIn = Ops.DataSize[0];
        *DataSizeOut = Ops.DataSize[1];
	}
	return ret;
}
/*****************************************************************************/
int32_t	ECAT_SetPdoInToOutCoeff(uint16_t DeviceNo, uint16_t ChannelNo, float gain, float offset)
{
	ecat_pdo_i2o_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
    Ops.gain = gain;
    Ops.offset = offset;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_PDO_I2O_COEFF,
							&Ops, sizeof(ecat_pdo_i2o_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t	ECAT_GetPdoInToOutCoeff(uint16_t DeviceNo, uint16_t ChannelNo, float *gain, float *offset)
{
	int32_t ret;
	ecat_pdo_i2o_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PDO_I2O_COEFF,
							&Ops, sizeof(ecat_pdo_i2o_ops_t), &Ops, sizeof(ecat_pdo_i2o_ops_t));
	if (ret == 0) {
        *gain = Ops.gain;
        *offset = Ops.offset;
	}
	return ret;
}
/*****************************************************************************/
int32_t	ECAT_SetPdoInToOutEnable(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t Enable)
{
	ecat_pdo_i2o_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
    Ops.Enable = Enable;
	return GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_SET_PDO_I2O_ENABLE,
							&Ops, sizeof(ecat_pdo_i2o_ops_t), NULL, 0);
}
/*****************************************************************************/
int32_t	ECAT_GetPdoInToOutEnable(uint16_t DeviceNo, uint16_t ChannelNo, uint16_t *Enable)
{
	int32_t ret;
	ecat_pdo_i2o_ops_t Ops;
    Ops.ChannelNo = ChannelNo;
	ret = GenericSendCommand(DeviceNo, 0, ECAT_IPC_CMD_GET_PDO_I2O_ENABLE,
							&Ops, sizeof(ecat_pdo_i2o_ops_t), &Ops, sizeof(ecat_pdo_i2o_ops_t));
	if (ret == 0) {
        *Enable = Ops.Enable;
	}
	return ret;
}
